from django.core.management.base import BaseCommand
from django.db import transaction

from arches_querysets.utils.tests import GraphTestCase


class Command(BaseCommand):
    def handle(self, *args, **options):
        self.stdout.write("Creating test data...")
        with transaction.atomic():
            GraphTestCase.setUpTestData()
        self.stdout.write("Finished!")



import datetime
import uuid

from django.db.models import F, OuterRef, prefetch_related_objects
from django.test import TestCase

from arches import VERSION as arches_version
from arches.app.models.graph import Graph
from arches.app.models.models import (
    CardModel,
    CardXNodeXWidget,
    Concept,
    DDataType,
    Edge,
    Node,
    NodeGroup,
    ResourceInstance,
    ResourceXResource,
    TileModel,
)

from arches_querysets.datatypes.datatypes import DataTypeFactory
from arches_querysets.models import GraphWithPrefetching, ResourceTileTree


class GraphTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.datatype_factory = DataTypeFactory()  # custom!
        cls.create_graph()
        cls.create_nodegroups_and_grouping_nodes()
        cls.create_data_collecting_nodes()
        cls.create_child_nodegroups()
        cls.create_edges()
        cls.create_cards()
        cls.create_widgets()
        cls.add_default_values_for_widgets()
        cls.create_resources()
        cls.create_tiles_with_data()
        cls.create_tiles_with_none()
        cls.create_child_tiles()
        cls.create_relations()

        graph_proxy = Graph.objects.get(pk=cls.graph.pk)
        # arches_version==9.0.0
        if arches_version < (8, 0):
            graph_proxy.refresh_from_database()
            # Repair parent nodegroup link broken by get_nodegroups()!
            for node in graph_proxy.nodes.values():
                if node.nodegroup == cls.nodegroup_1_child:
                    if node.nodegroup.parentnodegroup != cls.nodegroup_1:
                        node.nodegroup.parentnodegroup = cls.nodegroup_1
                        node.nodegroup.save()
        graph_proxy.publish(user=None)
        cls.graph.publication = graph_proxy.publication
        for resource in [cls.resource_42, cls.resource_none]:
            resource.graph_publication = cls.graph.publication
        # resource.save() might complain about differing publications, but
        # we're trying to sync the publications here, so sidestep it.
        ResourceInstance.objects.annotate(
            graph_pub_id=Graph.objects.filter(resourceinstance=OuterRef("pk")).values(
                "publication"
            )[:1]
        ).update(graph_publication=F("graph_pub_id"))

    @classmethod
    def create_graph(cls):
        cls.graph = GraphWithPrefetching.objects.create_graph(
            name="Datatype Lookups", is_resource=True
        )
        # arches_version==9.0.0
        if arches_version >= (8, 0):
            cls.graph.is_active = True
            cls.graph.save()
        cls.root_node = cls.graph.node_set.get(istopnode=True)

    @classmethod
    def create_nodegroup(cls, alias, cardinality, parent_nodegroup=None):
        grouping_node = Node(
            graph=cls.graph, alias=alias, istopnode=False, datatype="semantic"
        )
        nodegroup = NodeGroup.objects.create(
            pk=grouping_node.pk,
            cardinality=cardinality,
            parentnodegroup=parent_nodegroup,
        )
        # arches_version==9.0.0
        if arches_version >= (8, 0):
            nodegroup.grouping_node = grouping_node
            nodegroup.save()
        grouping_node.nodegroup = nodegroup
        grouping_node.save()
        return nodegroup, grouping_node

    @classmethod
    def create_nodegroups_and_grouping_nodes(cls):
        cls.nodegroup_1, cls.grouping_node_1 = cls.create_nodegroup("datatypes_1", "1")
        cls.nodegroup_n, cls.grouping_node_n = cls.create_nodegroup("datatypes_n", "n")

    @classmethod
    def create_data_collecting_nodes(cls):
        cls.datatypes = DDataType.objects.exclude(datatype="semantic").select_related(
            "defaultwidget"
        )
        cls.data_nodes_1 = [
            Node(
                datatype=datatype.pk,
                alias=datatype.pk.replace("-", "_") + "_alias",
                name=datatype.pk,
                istopnode=False,
                nodegroup=cls.nodegroup_1,
                graph=cls.graph,
                config=datatype.defaultconfig,
            )
            for datatype in cls.datatypes
        ]
        cls.data_nodes_n = [
            Node(
                datatype=datatype.pk,
                alias=datatype.pk.replace("-", "_") + "_alias_n",
                name=datatype.pk + "-n",
                istopnode=False,
                nodegroup=cls.nodegroup_n,
                graph=cls.graph,
                config=datatype.defaultconfig,
            )
            for datatype in cls.datatypes
        ]
        cls.data_nodes = Node.objects.bulk_create(cls.data_nodes_1 + cls.data_nodes_n)
        prefetch_related_objects(cls.data_nodes, "nodegroup")

        cls.setNodesOnClass()

        cls.node_value_node_1.config["nodeid"] = str(cls.date_node_1.pk)
        cls.node_value_node_1.save()
        cls.node_value_node_n.config["nodeid"] = str(cls.date_node_n.pk)
        cls.node_value_node_n.save()

    @classmethod
    def setNodesOnClass(cls):
        """Set each node as an attribute, e.g. self.string_node_n"""
        for node in cls.data_nodes:
            attname = node.datatype.replace("-", "_") + "_node"
            attname += "_1" if node.nodegroup.cardinality == "1" else "_n"
            attname += "_child" if node.nodegroup.parentnodegroup_id else ""
            setattr(cls, attname, node)

    @classmethod
    def create_edges(cls):
        def get_node_to_append_to(node):
            if node.pk == node.nodegroup_id:
                if node.nodegroup.parentnodegroup_id:
                    if node == cls.nodegroup_1_child:
                        return cls.grouping_node_1
                    else:
                        return cls.grouping_node_n
                return cls.root_node
            if node.nodegroup == cls.nodegroup_1:
                return cls.grouping_node_1
            if node.nodegroup == cls.nodegroup_n:
                return cls.grouping_node_n
            if node.nodegroup == cls.nodegroup_1_child:
                return cls.grouping_node_1_child
            if node.nodegroup == cls.nodegroup_n_child:
                return cls.grouping_node_n_child
            raise ValueError

        edges = [
            Edge(
                domainnode=get_node_to_append_to(node),
                rangenode=node,
                ontologyproperty="",
                graph=cls.graph,
            )
            for node in cls.graph.node_set.exclude(pk=cls.root_node.pk).select_related(
                "nodegroup"
            )
        ]
        cls.edges = Edge.objects.bulk_create(edges)

    @classmethod
    def create_cards(cls):
        nodegroups = [cls.nodegroup_1, cls.nodegroup_n]
        nodegroups.extend([cls.nodegroup_1_child, cls.nodegroup_n_child])
        cards = [
            CardModel(
                graph=cls.graph,
                nodegroup=nodegroup,
            )
            for nodegroup in nodegroups
        ]
        cards = CardModel.objects.bulk_create(cards)

    @classmethod
    def create_widgets(cls):
        node_widgets = [
            CardXNodeXWidget(
                node=node,
                widget_id=cls.find_default_widget_id(node, cls.datatypes),
                card=node.nodegroup.cardmodel_set.all()[0],
            )
            for node in cls.data_nodes
        ]
        CardXNodeXWidget.objects.bulk_create(node_widgets)

    @classmethod
    def add_default_values_for_widgets(cls):
        node_widgets = CardXNodeXWidget.objects.filter(
            node__graph=cls.graph,
        ).select_related("node")
        cls.default_vals_by_nodeid = {}
        cls.default_vals_by_datatype = {
            "non-localized-string": "The answer to life, the universe, and everything.",
            "string": {
                "en": {
                    "value": "The answer to life, the universe, and everything.",
                    "direction": "ltr",
                }
            },
            "url": {"url": "http://arthurdent.com", "url_label": ""},
            "boolean": False,
            "number": 7,
            "date": (
                "1979-10-12"
                if arches_version > (8, 0)
                else "1979-10-12T00:00:00.000-05:00"
            ),
            # "resource-instance": None,
            # "resource-instance-list": [],
            # "concept": None,
            # "concept-list": [],
        }
        for widget in node_widgets:
            if widget.node.datatype in cls.default_vals_by_datatype:
                widget.config["defaultValue"] = cls.default_vals_by_datatype[
                    widget.node.datatype
                ]
                cls.default_vals_by_nodeid[str(widget.node.pk)] = (
                    cls.default_vals_by_datatype[widget.node.datatype]
                )
            else:
                for datatype in cls.datatypes:
                    if datatype.pk == widget.node.datatype:
                        nodeid = str(widget.node.pk)
                        config = datatype.defaultwidget.defaultconfig
                        cls.default_vals_by_nodeid[nodeid] = config.get("defaultValue")
                        break
                else:
                    raise RuntimeError("Missing datatype")
            # arches_version==9.0.0
            if arches_version < (8, 0):
                widget.save()
        # arches_version==9.0.0
        if arches_version >= (8, 0):
            CardXNodeXWidget.objects.bulk_update(node_widgets, ["config"])

    @classmethod
    def create_resources(cls):
        if arches_version >= (8, 0):
            from arches.app.models.models import ResourceInstanceLifecycleState

            state = ResourceInstanceLifecycleState.objects.first()

        resource_with_data_kwargs = {
            "graph": cls.graph,
            "descriptors": {"en": {"name": "Resource referencing 42"}},
            "name": "Resource referencing 42",
            "graph_publication_id": cls.graph.publication_id,
        }
        resource_without_data_kwargs = {
            "graph": cls.graph,
            "descriptors": {"en": {"name": "Resource referencing None"}},
            "name": "Resource referencing None",
            "graph_publication_id": cls.graph.publication_id,
        }

        # arches_version==9.0.0
        if arches_version >= (8, 0):
            resource_with_data_kwargs["resource_instance_lifecycle_state"] = state
            resource_without_data_kwargs["resource_instance_lifecycle_state"] = state

        cls.resource_42 = ResourceInstance.objects.create(**resource_with_data_kwargs)
        cls.resource_none = ResourceInstance.objects.create(
            **resource_without_data_kwargs
        )

    @classmethod
    def create_tiles_with_data(cls):
        ri_dt = cls.datatype_factory.get_instance("resource-instance")
        ri_list_dt = cls.datatype_factory.get_instance("resource-instance-list")

        cls.concept = Concept.objects.get(pk="00000000-0000-0000-0000-000000000001")
        cls.concept_value = cls.concept.value_set.get()

        cls.cardinality_1_tile = TileModel.objects.create(
            nodegroup=cls.nodegroup_1,
            resourceinstance=cls.resource_42,
            data={},
        )
        cls.cardinality_n_tile = TileModel.objects.create(
            nodegroup=cls.nodegroup_n,
            resourceinstance=cls.resource_42,
            data={},
        )

        cls.sample_data_1 = {
            "boolean": True,
            "number": 42,
            "non-localized-string": "forty-two",
            "string": {
                "en": {
                    "value": "forty-two",
                    "direction": "ltr",
                },
            },
            "url": {
                "url": "http://www.42.com/",
                "url_label": "42.com",
            },
            "date": "2042-04-02",
            "resource-instance": ri_dt.transform_value_for_tile(cls.resource_42),
            "resource-instance-list": ri_list_dt.transform_value_for_tile(
                cls.resource_42
            ),
            "concept": str(cls.concept_value.pk),
            "concept-list": [str(cls.concept_value.pk)],
            "node-value": str(cls.cardinality_1_tile.pk),
            "file-list": [
                {
                    "url": "http://www.archesproject.org/blog/static/42.png",
                    "name": "42_accessibility_improvements.png",
                    "size": 2042,
                    "type": "image/png",
                    "index": 0,
                    "title": {
                        "en": {
                            "value": "42 Accessibility Improvements",
                            "direction": "ltr",
                        }
                    },
                    "width": 2042,
                    "height": 2042,
                    "status": "added",
                    "altText": {
                        "en": {
                            "value": "Illustration of recent accessibility improvements",
                            "direction": "ltr",
                        }
                    },
                    "content": f"blob:http://localhost:8000/{uuid.uuid4()}",
                    "file_id": str(uuid.uuid4()),
                    "accepted": True,
                    "attribution": {"en": {"value": "Arches", "direction": "ltr"}},
                    "description": {
                        "en": {
                            "value": "Recent versions of arches have 42 improved accessibility characteristics.",
                            "direction": "ltr",
                        }
                    },
                    "lastModified": 1723503486969,
                }
            ],
            # TODO: geojson-feature-collection
            # TODO: domain-value{-list}
            # TODO(maybe): edtf
            # TODO(maybe): annotation
        }
        cls.sample_data_n = {
            **cls.sample_data_1,
            "node-value": str(cls.cardinality_n_tile.pk),
        }

        cls.cardinality_1_tile.data = {
            str(node.pk): (
                cls.sample_data_1[node.datatype]
                if node.datatype in cls.sample_data_1
                else None
            )
            for node in cls.data_nodes_1
        }
        cls.cardinality_1_tile.save()
        cls.cardinality_n_tile.data = {
            str(node.pk): (
                cls.sample_data_n[node.datatype]
                if node.datatype in cls.sample_data_n
                else None
            )
            for node in cls.data_nodes_n
        }
        cls.cardinality_n_tile.save()

    @classmethod
    def create_tiles_with_none(cls):
        cls.cardinality_1_tile_none = TileModel.objects.create(
            nodegroup=cls.nodegroup_1,
            resourceinstance=cls.resource_none,
            data={},
        )
        cls.cardinality_n_tile_none = TileModel.objects.create(
            nodegroup=cls.nodegroup_n,
            resourceinstance=cls.resource_none,
            data={},
        )

    @classmethod
    def create_relations(cls):
        # arches_version==9.0.0
        if arches_version < (8, 0):
            from_resource_attr = "resourceinstanceidto"
            to_resource_attr = "resourceinstanceidfrom"
            from_graph_attr = "resourceinstancefrom_graphid"
            to_graph_attr = "resourceinstanceto_graphid"
            tile_attr = "tileid"
            node_attr = "nodeid"
        else:
            from_resource_attr = "from_resource"
            to_resource_attr = "to_resource"
            from_graph_attr = "from_resource_graph"
            to_graph_attr = "to_resource_graph"
            tile_attr = "tile"
            node_attr = "node"
        rxrs = [
            ResourceXResource(
                **{
                    from_resource_attr: cls.resource_42,
                    to_resource_attr: cls.resource_42,
                    from_graph_attr: cls.graph,
                    to_graph_attr: cls.graph,
                    tile_attr: cls.cardinality_1_tile,
                    node_attr: cls.resource_instance_node_1,
                }
            ),
            ResourceXResource(
                **{
                    from_resource_attr: cls.resource_42,
                    to_resource_attr: cls.resource_42,
                    from_graph_attr: cls.graph,
                    to_graph_attr: cls.graph,
                    tile_attr: cls.cardinality_n_tile,
                    node_attr: cls.resource_instance_node_n,
                }
            ),
            ResourceXResource(
                **{
                    from_resource_attr: cls.resource_42,
                    to_resource_attr: cls.resource_42,
                    from_graph_attr: cls.graph,
                    to_graph_attr: cls.graph,
                    tile_attr: cls.cardinality_1_tile,
                    node_attr: cls.resource_instance_list_node_1,
                }
            ),
            ResourceXResource(
                **{
                    from_resource_attr: cls.resource_42,
                    to_resource_attr: cls.resource_42,
                    from_graph_attr: cls.graph,
                    to_graph_attr: cls.graph,
                    tile_attr: cls.cardinality_n_tile,
                    node_attr: cls.resource_instance_list_node_n,
                }
            ),
        ]
        for rxr in rxrs:
            rxr.created = datetime.datetime.now()
            rxr.modified = datetime.datetime.now()
        ResourceXResource.objects.bulk_create(rxrs)

    @classmethod
    def create_child_nodegroups(cls):
        cls.nodegroup_1_child, cls.grouping_node_1_child = cls.create_nodegroup(
            "datatypes_1_child", "1", parent_nodegroup=cls.nodegroup_1
        )
        cls.nodegroup_n_child, cls.grouping_node_n_child = cls.create_nodegroup(
            "datatypes_n_child", "n", parent_nodegroup=cls.nodegroup_n
        )

        # Clone nodes.
        for node in cls.data_nodes:
            node.pk = uuid.uuid4()
            node.name = node.name + "-child"
            node.alias = node.alias + "_child"
            node.nodegroup = (
                cls.nodegroup_1_child
                if node.nodegroup == cls.nodegroup_1
                else cls.nodegroup_n_child
            )

        Node.objects.bulk_create(cls.data_nodes)

        # Refresh class attributes.
        cls.data_nodes = cls.graph.node_set.exclude(datatype="semantic").select_related(
            "nodegroup"
        )
        cls.data_nodes_1 = cls.graph.node_set.filter(nodegroup=cls.nodegroup_1).exclude(
            datatype="semantic"
        )
        cls.data_nodes_n = cls.graph.node_set.filter(nodegroup=cls.nodegroup_n).exclude(
            datatype="semantic"
        )
        cls.setNodesOnClass()

    @classmethod
    def create_child_tiles(cls):
        cls.cardinality_1_child_tile = TileModel.objects.create(
            nodegroup=cls.nodegroup_1_child,
            resourceinstance=cls.resource_42,
            data={},
            parenttile=cls.cardinality_1_tile,
        )
        cls.cardinality_n_child_tile = TileModel.objects.create(
            nodegroup=cls.nodegroup_n_child,
            resourceinstance=cls.resource_42,
            data={},
            parenttile=cls.cardinality_n_tile,
        )
        cls.cardinality_1_child_tile_none = TileModel.objects.create(
            nodegroup=cls.nodegroup_1_child,
            resourceinstance=cls.resource_none,
            data={},
            parenttile=cls.cardinality_1_tile_none,
        )
        cls.cardinality_n_child_tile_none = TileModel.objects.create(
            nodegroup=cls.nodegroup_n_child,
            resourceinstance=cls.resource_none,
            data={},
            parenttile=cls.cardinality_n_tile_none,
        )

        # Create data for the child non-localized-string node only.
        # TileModel.save() will initialize the other nodes to None.
        cls.non_localized_string_child_node = Node.objects.get(
            alias="non_localized_string_alias_child"
        )
        cls.non_localized_string_child_node_n = Node.objects.get(
            alias="non_localized_string_alias_n_child"
        )
        cls.cardinality_1_child_tile.data = {
            str(cls.non_localized_string_child_node.pk): "child-1-value",
        }
        cls.cardinality_1_child_tile.save()
        cls.cardinality_n_child_tile.data = {
            str(cls.non_localized_string_child_node_n.pk): "child-n-value",
        }
        cls.cardinality_n_child_tile.save()

        cls.resource_42 = ResourceTileTree.get_tiles(
            "datatype_lookups", as_representation=True
        ).get(pk=cls.resource_42.pk)
        cls.resource_none = ResourceTileTree.get_tiles(
            "datatype_lookups", as_representation=True
        ).get(pk=cls.resource_none.pk)

    @classmethod
    def find_default_widget_id(cls, node, datatypes):
        for datatype in datatypes:
            if node.datatype == datatype.pk:
                return datatype.defaultwidget_id
        return None



class NodeValueMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def to_representation(self, value):
        try:
            value["node_value"]
        except (TypeError, KeyError):
            return super().to_representation(value)
        return value

    def to_internal_value(self, data):
        try:
            node_value = data["node_value"]
        except (TypeError, KeyError):
            node_value = data
        # DRF's DateField doesn't handle None despite a few
        # close-but-no-cigar bug reports like:
        # https://github.com/encode/django-rest-framework/issues/4835
        if node_value is None:
            return None
        return super().to_internal_value(node_value)



from functools import partial
from itertools import chain

from django.core.exceptions import ValidationError as DjangoValidationError
from django.db.models import Q
from django.db.models.constants import LOOKUP_SEP
from django.utils.translation import gettext as _
from rest_framework.exceptions import PermissionDenied, ValidationError
from rest_framework.metadata import SimpleMetadata
from rest_framework.settings import api_settings

from arches import VERSION as arches_version
from arches.app.models.models import Node, ResourceInstance, TileModel
from arches.app.utils.permission_backend import (
    user_can_delete_resource,
    user_can_edit_resource,
    user_can_read_resource,
)
from arches.app.utils.string_utils import str_to_bool

from arches_querysets.models import TileTree
from arches_querysets.utils.models import ensure_request


class MetadataWithWidgetConfig(SimpleMetadata):
    def get_field_info(self, field):
        return {
            **super().get_field_info(field),
            "key": field.style.get("alias"),
            "initial": (
                None if field.initial is field.default_empty_html else field.initial
            ),
            "visible": field.style.get("visible", True),
            "datatype": field.style.get("datatype"),
            "widget_config": field.style.get("widget_config", {}),
            "sortorder": field.style.get("sortorder", 0),
        }


class ArchesModelAPIMixin:
    metadata_class = MetadataWithWidgetConfig

    def setup(self, request, *args, **kwargs):
        options = self.serializer_class.Meta
        self.graph_slug = options.graph_slug or kwargs.get("graph")
        self.nodegroup_alias = kwargs.get("nodegroup_alias")
        self.nodegroup_alias_lookup = {}

        if issubclass(options.model, TileModel):
            self.nodegroup_alias = options.root_node or self.nodegroup_alias

        if resource_ids := request.GET.get("resource_ids"):
            self.resource_ids = resource_ids.split(",")
        elif issubclass(options.model, ResourceInstance) and (pk := kwargs.get("pk")):
            self.resource_ids = [pk]
        else:
            self.resource_ids = None
        self.fill_blanks = request.GET.get("fill_blanks", "f").lower().startswith("t")

        # Graph nodes are supplied by the underlying queryset, either by
        # get_object() for detail views or filter_queryset() for list views.
        # creates don't have to worry about getting the same graph objects
        # from the retrieval operation: falls back to NodeFetcherMixin._find_graph_nodes()
        self.graph_nodes = Node.objects.none()

        return super().setup(request, *args, **kwargs)

    def get_queryset(self):
        options = self.serializer_class.Meta
        try:
            if issubclass(options.model, ResourceInstance):
                qs = options.model.get_tiles(
                    self.graph_slug,
                    resource_ids=self.resource_ids,
                    as_representation=True,
                ).select_related("graph")
                if arches_version >= (8, 0):
                    qs = qs.select_related("resource_instance_lifecycle_state")
            elif issubclass(options.model, TileModel):
                qs = options.model.get_tiles(
                    self.graph_slug,
                    self.nodegroup_alias,
                    as_representation=True,
                    resource_ids=self.resource_ids,
                ).select_related("nodegroup", "resourceinstance__graph")
                if self.resource_ids:
                    qs = qs.filter(resourceinstance__in=self.resource_ids)
            else:  # pragma: no cover
                raise NotImplementedError
        except ValueError:
            msg = _("No nodes found for graph slug: %s") % self.graph_slug
            raise ValidationError({api_settings.NON_FIELD_ERRORS_KEY: msg})

        return qs

    def get_serializer_context(self):
        context = super().get_serializer_context()
        ret = {
            **context,
            "request": ensure_request(context["request"]),
            "graph_slug": self.graph_slug,
            "graph_nodes": self.graph_nodes or self._find_graph_nodes(),
            "nodegroup_alias": self.nodegroup_alias,
            "fill_blanks": self.fill_blanks,
        }
        if not ret.get("nodegroup_alias_lookup"):
            ret["nodegroup_alias_lookup"] = {
                node.pk: node.alias
                for node in ret["graph_nodes"]
                if node.pk == node.nodegroup_id
            }
        return ret

    def _find_graph_nodes(self):
        node_filters = Q(graph__slug=self.graph_slug, nodegroup__isnull=False)
        children = "nodegroup_set"
        # arches_version==9.0.0
        if arches_version >= (8, 0):
            node_filters &= Q(source_identifier=None)
            children = "children"

        return (
            Node.objects.filter(node_filters)
            .select_related("nodegroup")
            .prefetch_related(
                "nodegroup__node_set",
                "nodegroup__cardmodel_set",
                f"nodegroup__{children}",
                "cardxnodexwidget_set",
            )
        )

    def get_object(self, permission_callable=None, fill_blanks=False):
        ret = super().get_object()
        if isinstance(ret, TileTree):
            self.graph_nodes = ret.resourceinstance.graph.node_set.all()
        else:
            self.graph_nodes = ret.graph.node_set.all()
        self.nodegroup_alias_lookup = {
            node.pk: node.alias
            for node in self.graph_nodes
            if node.pk == node.nodegroup_id
        }

        options = self.serializer_class.Meta
        if issubclass(options.model, ResourceInstance):
            # arches_version==9.0.0
            if arches_version >= (8, 0):
                permission_kwargs = {"user": self.request.user, "resource": ret}
            else:
                permission_kwargs = {"user": self.request.user, "resourceid": ret.pk}
            if not self.graph_slug:
                # Resource results for heterogenous graphs are not supported.
                self.graph_slug = ret.graph.slug
        else:
            permission_kwargs = {
                "user": self.request.user,
                "resourceid": ret.resourceinstance_id,
            }
            if not self.graph_slug:
                self.graph_slug = ret.resourceinstance.graph.slug
        if permission_callable and not permission_callable(**permission_kwargs):
            # Not 404, see https://github.com/archesproject/arches/issues/11563
            raise PermissionDenied

        # Freeze some keyword arguments to the model save() method.
        is_partial_update = self.request.method == "PATCH"
        ret.save = partial(ret.save, request=self.request, partial=is_partial_update)

        if fill_blanks:
            ret.fill_blanks()

        return ret

    def filter_queryset(self, queryset):
        if queryset._hints.get("graph_query") is not None:
            self.graph_nodes = next(
                iter(queryset._hints.get("graph_query"))
            ).node_set.all()
            self.nodegroup_alias_lookup = {
                node.pk: node.alias
                for node in self.graph_nodes
                if node.pk == node.nodegroup_id
            }

        # Parse ORM lookpus from query params starting with "aliased_data__"
        # This is a quick-n-dirty riff on https://github.com/miki725/django-url-filter
        # minus some features like negation (`!=`)
        data_filters = {
            param.split("aliased_data__", maxsplit=1)[1]: (
                str_to_bool(value) if param.endswith("__isnull") else value
            )
            for param, value in self.request.GET.items()
            if param.startswith("aliased_data__")
        }
        node_aliases = [n.alias for n in self.graph_nodes]
        for filter_value in data_filters:
            first_filter_term = filter_value.split(LOOKUP_SEP, maxsplit=1)[0]
            if first_filter_term not in node_aliases:
                msg = _("Invalid filter param: %s") % first_filter_term
                raise ValidationError(msg)
        queryset = queryset.filter(**data_filters)

        return super().filter_queryset(queryset)

    def create(self, request, *args, **kwargs):
        self.get_object = partial(
            self.get_object,
            permission_callable=user_can_edit_resource,
            fill_blanks=self.fill_blanks,
        )
        return super().create(request, *args, **kwargs)

    def retrieve(self, request, *args, **kwargs):
        self.get_object = partial(
            self.get_object,
            permission_callable=user_can_read_resource,
            fill_blanks=self.fill_blanks,
        )
        return super().retrieve(request, *args, **kwargs)

    def update(self, request, *args, **kwargs):
        self.get_object = partial(
            self.get_object,
            permission_callable=user_can_edit_resource,
            fill_blanks=self.fill_blanks,
        )
        return super().update(request, *args, **kwargs)

    def destroy(self, request, *args, **kwargs):
        self.get_object = partial(
            self.get_object,
            permission_callable=user_can_delete_resource,
        )
        return super().destroy(request, *args, **kwargs)

    def validate_tile_data_and_save(self, serializer):
        """Re-raise ValidationError as DRF ValidationError.

        In 3.0 (2014), DRF decided to stop full_clean()'ing before save(),
        which divorces DRF validation needs from model logic needing to
        support the Django admin or similar ModelFormish patterns.
        The stated reasons were:
            - to avoid calling into big & scary full_clean().
            - to force expressing validation logic outside of models.
        but adhering to that second point would be difficult in light of
        how dynamically these fields are constructed.

        Discussion:
        https://github.com/encode/django-rest-framework/discussions/7850
        """
        try:
            serializer.save()
        except DjangoValidationError as django_error:
            flattened_errors = self.flatten_validation_errors(django_error)
            if isinstance(flattened_errors, dict):
                errors = flattened_errors
            else:
                errors = {api_settings.NON_FIELD_ERRORS_KEY: flattened_errors}
            raise ValidationError(errors) from django_error

    def perform_create(self, serializer):
        self.validate_tile_data_and_save(serializer)

    def perform_update(self, serializer):
        self.validate_tile_data_and_save(serializer)

    @staticmethod
    def flatten_validation_errors(error):
        """DRF's ValidationError doesn't really handle nesting, so unpack
        one level."""
        if hasattr(error, "error_dict"):
            return {
                k: (
                    list(chain.from_iterable(inner.messages for inner in v))
                    if all(isinstance(inner, DjangoValidationError) for inner in v)
                    else v
                )
                for k, v in error.error_dict.items()
            }
        else:
            return error.messages



import logging
import uuid
from collections import defaultdict
from operator import attrgetter

from django.core.exceptions import ValidationError
from django.db import ProgrammingError, transaction
from django.db.models import F, Q, OuterRef, Exists
from django.utils.translation import get_language, gettext as _

from arches import VERSION as arches_version
from arches.app.models.models import (
    CardXNodeXWidget,
    EditLog,
    Language,
    Node,
    ResourceInstance,
    TileModel,
)
from arches.app.models.resource import Resource
from arches.app.models.tile import Tile, TileValidationError

from arches_querysets.datatypes.datatypes import DataTypeFactory
from arches_querysets.utils.models import (
    field_attnames,
    get_nodegroups_here_and_below,
    pop_arches_model_kwargs,
)

logger = logging.getLogger(__name__)

NOT_PROVIDED = object()


class TileTreeOperation:
    def __init__(self, *, entry, request, partial=True, save_kwargs=None):
        self.to_insert = set()
        self.to_update = set()
        self.to_delete = set()
        self.errors_by_node_alias = defaultdict(list)
        self.entry = entry  # resource or tile
        self.datatype_factory = DataTypeFactory()
        self.languages = Language.objects.all()
        self.request = request
        self.partial = partial
        self.save_kwargs = save_kwargs or {}
        self.transaction_id = uuid.uuid4()
        # arches==9.0.0, replace these attributes by simply reading from
        # self.request.user.userprofile, which uses @cached_property as of Arches 8.1.
        # Store off these properties since they are expensive.
        self.editable_nodegroups: set[str] = (
            self.request.user.userprofile.editable_nodegroups
        )
        self.deletable_nodegroups: set[str] = (
            self.request.user.userprofile.deletable_nodegroups
        )

        if (self.partial and self.request.method == "PUT") or (
            not self.partial and self.request.method == "PATCH"
        ):
            raise ValueError(
                f"Expected partial={not self.partial} for {self.request.method} request."
            )

        if isinstance(entry, TileModel):
            self.resourceid = self.entry.resourceinstance_id
            self.nodegroups = get_nodegroups_here_and_below(self.entry.nodegroup)
            existing_tiles = (
                entry.__class__.objects.filter(
                    resourceinstance_id=self.resourceid,
                    nodegroup_id__in=[ng.pk for ng in self.nodegroups],
                )
                .select_related("nodegroup")
                .order_by("sortorder")
            )
            # arches_version==9.0.0
            if arches_version < (8, 0):
                # Cannot supply this too early, as nodegroup might be included
                # with the request and already instantiated to a fresh object.
                grouping_node = [
                    node
                    for node in entry.nodegroup.node_set.all()[0].graph.node_set.all()
                    if node.pk == node.nodegroup_id
                ][0]
                entry.nodegroup.grouping_node = grouping_node
        else:
            self.resourceid = self.entry.pk
            self.nodegroups = []  # not necessary to populate.
            existing_tiles = getattr(self.entry, "_tile_trees", [])

        self.for_new_resource = self.resourceid is None
        if self.for_new_resource:
            self.resourceid = uuid.uuid4()
            if isinstance(self.entry, TileModel):
                self.entry.resourceinstance_id = self.resourceid

        if isinstance(self.entry, TileModel):
            if self.for_new_resource:
                self.graph = self.entry.nodegroup.node_set.all()[0].graph
            else:
                self.graph = self.entry.resourceinstance.graph
        else:
            self.graph = self.entry.graph

        self.grouping_nodes_by_nodegroup_id = self._get_grouping_node_lookup()
        self.existing_tiles_by_nodegroup_alias = defaultdict(list)
        for tile in existing_tiles:
            alias = tile.find_nodegroup_alias(self.grouping_nodes_by_nodegroup_id)
            self.existing_tiles_by_nodegroup_alias[alias].append(tile)

    def _get_grouping_node_lookup(self):
        filters = Q(pk=F("nodegroup_id"), graph__slug=self.graph.slug)
        # arches_version==9.0.0
        if arches_version >= (8, 0):
            filters &= Q(source_identifier=None)
            return {
                node.pk: node
                for node in Node.objects.filter(filters)
                .select_related("nodegroup__grouping_node__nodegroup__parentnodegroup")
                .prefetch_related("nodegroup__node_set__cardxnodexwidget_set")
            }
        else:
            ret = {
                node.pk: node
                for node in Node.objects.filter(filters)
                .select_related("nodegroup__parentnodegroup")
                .prefetch_related("nodegroup__node_set__cardxnodexwidget_set")
            }
            for node_id, node in ret.items():
                node.nodegroup.grouping_node = ret[node_id]
            return ret

    def validate_and_save_tiles(self):
        delete_missing_tiles = (
            self.request.GET.get("delete_missing_tiles", str(not self.partial)).lower()
            == "true"
        )
        self.validate(delete_missing_tiles=delete_missing_tiles)
        try:
            self._save()
        except ProgrammingError as e:
            if e.args and "excess_tiles" in e.args[0]:
                nodegroup_id = e.args[0].split("nodegroupid: ")[1].split(",")[0]
                uid = uuid.UUID(nodegroup_id)
                nodegroup_alias = self.grouping_nodes_by_nodegroup_id[uid].alias
                msg = _("Tile Cardinality Error")
                raise ValidationError({nodegroup_alias: msg}) from e
            raise
        self.after_update_all()

    def validate(self, delete_missing_tiles=False):
        """Move values from resource or tile to prefetched tiles, and validate.
        Raises ValidationError if new data fails datatype validation.

        HTTP PUT should request delete_missing_tiles=True to delete child tiles
        not in the payload.
        HTTP PATCH should request delete_missing_tiles=False (default).
        """
        original_tile_data_by_tile_id = {}
        incoming_tiles = set()
        if isinstance(self.entry, TileModel):
            self._update_tile(
                self.grouping_nodes_by_nodegroup_id[self.entry.nodegroup_id],
                None,
                original_tile_data_by_tile_id,
                incoming_tiles=incoming_tiles,
                delete_missing_tiles=delete_missing_tiles,
            )
        else:
            for grouping_node in self.grouping_nodes_by_nodegroup_id.values():
                if grouping_node.nodegroup.parentnodegroup_id:
                    continue
                self._update_tile(
                    grouping_node,
                    self.entry,
                    original_tile_data_by_tile_id,
                    incoming_tiles=incoming_tiles,
                    delete_missing_tiles=delete_missing_tiles,
                )

        if self.errors_by_node_alias:
            raise ValidationError(
                {
                    alias: ValidationError([e["message"] for e in errors])
                    for alias, errors in self.errors_by_node_alias.items()
                }
            )

    def _update_tile(
        self,
        grouping_node,
        container,
        original_tile_data_by_tile_id,
        incoming_tiles,
        delete_missing_tiles=False,
    ):
        if str(grouping_node.nodegroup_id) not in self.editable_nodegroups:
            # Currently also prevents deletes.
            return

        try:
            new_tiles = self._extract_incoming_tiles(container, grouping_node)
            incoming_tiles.update(new_tiles)
        except KeyError:
            if delete_missing_tiles:
                for existing_tile in self.existing_tiles_by_nodegroup_alias[
                    grouping_node.alias
                ]:
                    if str(existing_tile.nodegroup_id) in self.deletable_nodegroups:
                        self.to_delete.add(existing_tile)
            return

        existing_tiles = self.existing_tiles_by_nodegroup_alias[grouping_node.alias]
        if not existing_tiles:
            next_sort_order = 0
        else:
            next_sort_order = max(t.sortorder or 0 for t in existing_tiles) + 1

        if self.for_new_resource or (
            isinstance(self.entry, ResourceInstance) and self.entry._state.adding
        ):
            exclude = ["resourceinstance"]
        else:
            exclude = []

        to_insert = set()
        to_update = set()
        to_delete = set()

        for existing_tile, new_tile in self._pair_tiles(existing_tiles, new_tiles):
            if new_tile is NOT_PROVIDED:
                if (
                    delete_missing_tiles
                    and str(existing_tile.nodegroup_id) in self.deletable_nodegroups
                ):
                    to_delete.add(existing_tile)
                continue

            if existing_tile is NOT_PROVIDED:
                new_tile.nodegroup_id = grouping_node.nodegroup_id
                new_tile.resourceinstance_id = self.resourceid
                new_tile.sortorder = next_sort_order
                next_sort_order += 1
                for node in grouping_node.nodegroup.node_set.all():
                    if node.datatype != "semantic":
                        new_tile.data[str(node.pk)] = new_tile.get_default_value(node)
                new_tile._incoming_tile = new_tile
                if isinstance(container, TileModel):
                    new_tile.parenttile = container
                new_tile.full_clean(exclude=exclude)
                if new_tile.pk is None:
                    new_tile.pk = uuid.uuid4()
                to_insert.add(new_tile)
            else:
                original_tile_data_by_tile_id[existing_tile.pk] = {**existing_tile.data}
                existing_tile._incoming_tile = new_tile
                to_update.add(existing_tile)

        nodes = grouping_node.nodegroup.node_set.all()
        for tile in to_insert | to_update:
            # arches_version==9.0.0
            if arches_version >= (8, 0):
                children = tile.nodegroup.children.all()
            else:
                children = tile.nodegroup.nodegroup_set.all()
            for child_nodegroup in children:
                self._update_tile(
                    grouping_node=self.grouping_nodes_by_nodegroup_id[
                        child_nodegroup.pk
                    ],
                    container=tile._incoming_tile,
                    original_tile_data_by_tile_id=original_tile_data_by_tile_id,
                    incoming_tiles=incoming_tiles,
                    delete_missing_tiles=delete_missing_tiles,
                )
            self._validate_and_patch_incoming_values(tile, nodes=nodes)

            tile.set_missing_keys_to_none()

        for tile in list(to_update):
            # Remove no-op upserts.
            if (
                original_data := original_tile_data_by_tile_id.pop(tile.pk, None)
            ) and tile._tile_update_is_noop(original_data):
                to_update.remove(tile)

        self.to_insert |= to_insert
        self.to_update |= to_update
        self.to_delete |= to_delete

        for tile in incoming_tiles:
            # Need to remove tiles flagged for deletion if they are
            # being updated.  This can happen becuase of the way cardinality n tiles
            # are processed within separate _update_tile() calls.
            # print("DEBUG: Removing tile from delete set because it is being upserted, tile id", str(tile.pk))
            self.to_delete.discard(tile)

    def _extract_incoming_tiles(self, container, grouping_node):
        from arches_querysets.models import TileTree

        if container is None:
            aliased_data = self.entry.aliased_data
        elif isinstance(container, dict):
            aliased_data = container.get("aliased_data")
        else:
            aliased_data = container.aliased_data

        if container is None:
            new_tiles = [self.entry]
        elif isinstance(aliased_data, dict):
            new_tiles = aliased_data.get(grouping_node.alias, NOT_PROVIDED)
        else:
            new_tiles = getattr(aliased_data, grouping_node.alias, NOT_PROVIDED)
        if new_tiles is NOT_PROVIDED:
            raise KeyError(grouping_node.alias)

        if grouping_node.nodegroup.cardinality == "1":
            if new_tiles is None:
                new_tiles = []
            elif not isinstance(new_tiles, list):
                new_tiles = [new_tiles]

        if all(isinstance(tile, TileTree) for tile in new_tiles):
            new_tiles.sort(
                key=lambda tile_tree: (tile_tree.sortorder is None, tile_tree.sortorder)
            )
        else:
            parent_tile = container if isinstance(container, TileTree) else None
            new_tiles = [
                TileTree.deserialize(tile_dict, parent_tile=parent_tile)
                for tile_dict in new_tiles
            ]

        return new_tiles

    def _pair_tiles(self, existing_tiles, new_tiles):
        pairs = []
        matched_new_tiles = []
        for existing_tile in existing_tiles:
            for tile in new_tiles:
                if str(existing_tile.pk) == str(tile.pk):
                    pairs.append((existing_tile, tile))
                    matched_new_tiles.append(tile)
                    break
            else:
                pairs.append((existing_tile, NOT_PROVIDED))
        for new_tile in new_tiles:
            if new_tile not in matched_new_tiles:
                pairs.append((NOT_PROVIDED, new_tile))
        return pairs

    def _validate_and_patch_incoming_values(self, tile, *, nodes):
        """Validate data found on tile._incoming_tile and move it to tile.data.
        Update errors_by_node_alias in place."""
        from arches_querysets.models import AliasedData, TileTree

        for node in nodes:
            if node.datatype == "semantic":
                continue
            if isinstance(tile._incoming_tile, TileTree) and isinstance(
                tile._incoming_tile.aliased_data, AliasedData
            ):
                value_to_validate = getattr(
                    tile._incoming_tile.aliased_data, node.alias, NOT_PROVIDED
                )
            else:
                value_to_validate = tile._incoming_tile.aliased_data.get(
                    node.alias, NOT_PROVIDED
                )
            if value_to_validate is NOT_PROVIDED:
                if self.partial:
                    continue
                value_to_validate = TileTree.get_default_value(node)
                tile._incoming_tile.set_aliased_data(node, value_to_validate)
            if isinstance(value_to_validate, dict):
                value_to_validate = value_to_validate.get(
                    "node_value", value_to_validate
                )

            self._run_datatype_methods(tile, value_to_validate, node)

    def _run_datatype_methods(self, tile, value_to_validate, node):
        """
        Call datatype methods when merging value_to_validate into the tile.

        1. transform_value_for_tile() -- type coercion
        2. pre_structure_tile_data() -- insert missing dictionary keys
        3. clean() -- replace empty values
        4. validate() -- check business logic, don't mutate data
        5. pre_tile_save() -- run side effects

        TODO: move this to BaseDataType.full_clean()?
        https://github.com/archesproject/arches/issues/10851#issuecomment-2427305853
        """
        node_id_str = str(node.pk)
        datatype_instance = self.datatype_factory.get_instance(node.datatype)

        if value_to_validate is None:
            tile.data[node_id_str] = None
            return
        try:
            transformed = datatype_instance.transform_value_for_tile(
                value_to_validate,
                languages=self.languages,
                is_existing_tile=(not tile._state.adding),
                **node.config,
            )
        except ValueError:  # BooleanDataType raises.
            # validate() will handle.
            transformed = value_to_validate

        # Merge the incoming value.
        tile.data[node_id_str] = transformed

        datatype_instance.pre_structure_tile_data(
            tile, node_id_str, languages=self.languages
        )

        datatype_instance.clean(tile, node_id_str)

        if errors := datatype_instance.validate(
            transformed, node=node, request=self.request
        ):
            self.errors_by_node_alias[node.alias].extend(errors)

        try:
            datatype_instance.pre_tile_save(tile, node_id_str)
        except TypeError:  # GeoJSONDataType raises.
            self.errors_by_node_alias[node.alias].append(
                datatype_instance.create_error_message(
                    tile.data[node_id_str], None, None, None
                )
            )

    def _save(self):
        if self.for_new_resource:
            instance = Resource(pk=self.resourceid, graph=self.graph)
            instance.save(request=self.request, transaction_id=self.transaction_id)
            # if entry tile needs saving, already in self.to_update or self.to_insert

        # durable=True is a guard against any higher-level code trying to wrap in
        # another transaction. Ideally durable=True would be removed, and any
        # IntegrityErrors would not be ignored in after_update_all(), but we have
        # some Arches projects that test with "incomplete" datasets with data
        # integrity issues, so we can't do the natural thing (yet).
        try:
            self._perform_transaction()
        except:
            if self.for_new_resource:
                # Manually manage failures instead of using transaction.atomic(), see:
                # https://github.com/archesproject/arches/issues/12318
                # Don't want to run model delete() which *creates* edit log entries.
                EditLog.objects.filter(resourceinstanceid=instance.pk).delete()
                Resource.objects.filter(pk=instance.pk).delete()
            raise

    def _perform_transaction(self):
        from arches_querysets.models import ResourceTileTree

        # Instantiate proxy models for now, but TODO: expose this
        # functionality on vanilla models, and in bulk.
        tile_model_fields = Tile._meta.get_fields()
        upserts = self.to_insert | self.to_update
        insert_proxies = [
            # Instantiate TileProxyModel instances without aliased data.
            Tile(**(pop_arches_model_kwargs(vars(insert), tile_model_fields)[1]))
            for insert in self.to_insert
        ]
        update_proxies = list(
            Tile.objects.filter(pk__in=[tile.pk for tile in self.to_update])
            .select_related("resourceinstance__graph")
            .prefetch_related("nodegroup__cardmodel_set")
        )
        upsert_proxies = insert_proxies + update_proxies
        delete_proxies = Tile.objects.filter(
            pk__in=[tile.pk for tile in self.to_delete]
        )

        with transaction.atomic(durable=True):
            if not self.for_new_resource:
                if isinstance(self.entry, ResourceTileTree):
                    super(ResourceTileTree, self.entry).save(**self.save_kwargs)
                elif isinstance(self.entry, ResourceInstance):
                    self.entry.save(**self.save_kwargs)
            # Interact with the database in bulk as much as possible, but
            # run certain side effects from Tile.save() one-at-a-time until
            # proxy model methods can be refactored. Then run in bulk.
            for upsert_proxy, vanilla_instance in zip(
                sorted(upsert_proxies, key=attrgetter("pk")),
                sorted(upserts, key=attrgetter("pk")),
                strict=True,
            ):
                upsert_proxy._existing_data = upsert_proxy.data
                upsert_proxy._existing_provisionaledits = upsert_proxy.provisionaledits

                # Sync proxy instance fields.
                for field in field_attnames(vanilla_instance):
                    setattr(upsert_proxy, field, getattr(vanilla_instance, field))

                # Some functions expect to always drill into request.user
                # https://github.com/archesproject/arches/issues/8471
                try:
                    upsert_proxy._Tile__preSave(request=self.request)
                    upsert_proxy.check_for_missing_nodes()  # also runs clean()
                    upsert_proxy.check_for_constraint_violation()
                except TileValidationError as error:
                    if ":" in error.message:
                        self.parse_required_error(error)
                    raise ValidationError(error.message) from error
                (
                    oldprovisionalvalue,
                    newprovisionalvalue,
                    provisional_edit_log_details,
                ) = vanilla_instance._apply_provisional_edit(
                    upsert_proxy,
                    upsert_proxy._existing_data,
                    upsert_proxy._existing_provisionaledits,
                    user=self.request.user,
                )
                # Remember the values needed for the edit log updates later.
                upsert_proxy._oldprovisionalvalue = oldprovisionalvalue
                upsert_proxy._newprovisionalvalue = newprovisionalvalue
                upsert_proxy._provisional_edit_log_details = (
                    provisional_edit_log_details
                )
                upsert_proxy._existing_data = vanilla_instance.data

            for delete_proxy in delete_proxies:
                delete_proxy._Tile__preDelete(request=self.request)

            if self.to_insert:
                inserted = sorted(
                    TileModel.objects.bulk_create(self.to_insert), key=attrgetter("pk")
                )
                # Pay the cost of a second TileModel -> Tile transform until refactored.
                refreshed_insert_proxies = list(
                    Tile.objects.filter(pk__in=[t.pk for t in inserted]).order_by("pk")
                )
                for before, after in zip(
                    insert_proxies, refreshed_insert_proxies, strict=True
                ):
                    after._newprovisionalvalue = before._newprovisionalvalue
                    after._provisional_edit_log_details = (
                        before._provisional_edit_log_details
                    )
                upsert_proxies = refreshed_insert_proxies + update_proxies
            else:
                insert_proxies = Tile.objects.none()
            if self.to_update:
                TileModel.objects.bulk_update(
                    self.to_update,
                    # No updates to resource instance or nodegroup.
                    {"data", "parenttile", "provisionaledits", "sortorder"},
                )
            if self.to_delete:
                TileModel.objects.filter(pk__in=[t.pk for t in self.to_delete]).delete()

            for upsert_tile in upserts:
                grouping_node = self.grouping_nodes_by_nodegroup_id[
                    upsert_tile.nodegroup_id
                ]
                for node in grouping_node.nodegroup.node_set.all():
                    datatype = self.datatype_factory.get_instance(node.datatype)
                    datatype.post_tile_save(
                        upsert_tile, str(node.pk), request=self.request
                    )

            for upsert_proxy in upsert_proxies:
                upsert_proxy._Tile__postSave()

            # Save edits: could be done in bulk once above side effects are un-proxied.
            for insert_proxy in insert_proxies:
                insert_proxy.save_edit(
                    user=self.request.user,
                    edit_type="tile create",
                    old_value={},
                    new_value=insert_proxy.data,
                    newprovisionalvalue=insert_proxy._newprovisionalvalue,
                    provisional_edit_log_details=insert_proxy._provisional_edit_log_details,
                    transaction_id=self.transaction_id,
                    new_resource_created=False,  # resource create edit log already saved
                    note=None,
                )
            for update_proxy in update_proxies:
                update_proxy.save_edit(
                    user=self.request.user,
                    edit_type="tile edit",
                    old_value=update_proxy._existing_data,
                    new_value=update_proxy.data,
                    newprovisionalvalue=update_proxy._newprovisionalvalue,
                    oldprovisionalvalue=update_proxy._oldprovisionalvalue,
                    provisional_edit_log_details=update_proxy._provisional_edit_log_details,
                    transaction_id=self.transaction_id,
                )
            for delete_proxy in delete_proxies:
                delete_proxy.save_edit(
                    user=self.request.user,
                    edit_type="tile delete",
                    old_value=delete_proxy.data,
                    provisional_edit_log_details=None,
                    transaction_id=self.transaction_id,
                )

    def after_update_all(self):
        for datatype in self.datatype_factory.datatype_instances.values():
            try:
                if (
                    datatype.__class__.__name__ == "GeojsonFeatureCollectionDataType"
                    and self.resourceid
                ):
                    nodes_exist_subquery = Node.objects.filter(
                        nodegroup_id=OuterRef("nodegroup_id"),
                        datatype="geojson-feature-collection",
                    )
                    tiles = (
                        TileModel.objects.filter(
                            resourceinstance_id=self.resourceid,
                        )
                        .annotate(has_geojson_node=Exists(nodes_exist_subquery))
                        .filter(has_geojson_node=True)
                        .all()
                    )
                    for tile in tiles:
                        datatype.after_update_all(tile)
                else:
                    datatype.after_update_all()
            except:
                # This wide catch can leave the DB in an unusable state, so not only is
                # this the *last* operation, but durable=True on the transaction.
                # https://github.com/archesproject/arches/issues/12318
                logger.error(
                    f"Error in {datatype.__class__.__name__}.after_update_all():",
                    exc_info=True,
                )
                continue

    def parse_required_error(self, error):
        """
        This is a make-do attempt to parse the TileValidationError,
        which should raise something more discrete than a localized
        "This card requires values for the following: "
        Some translations of that string do not include whitespace
        after the colon, so be sure to only strip the space, not split.
        """
        error_names = [name.strip() for name in error.message.split(":")[1].split(", ")]
        aliases = []

        from arches_querysets.models import TileTree

        if isinstance(self.entry, TileTree):
            nodes = self.entry.resourceinstance.graph.node_set.all()
        else:
            nodes = self.entry.graph.node_set.all()
        for widget_label_or_node_name in error_names:
            widget = CardXNodeXWidget.objects.filter(
                node__in=nodes,
                # Awkward due to I18n_JSON
                label__contains={get_language(): widget_label_or_node_name},
            ).first()
            if widget:
                node = widget.node
            else:
                node = Node.objects.filter(
                    pk__in=nodes,
                    name=widget_label_or_node_name,
                ).first()
            if node:
                aliases.append(node.alias)
            else:
                msg = "Unable to backsolve error name: %s"
                logger.error(msg, widget_label_or_node_name)
                raise ValidationError(error.message) from error
        raise ValidationError({alias: error.message for alias in aliases}) from error



from django.conf import settings
from rest_framework.pagination import LimitOffsetPagination


class ArchesLimitOffsetPagination(LimitOffsetPagination):
    default_limit = settings.API_MAX_PAGE_SIZE



import ast
import json
import re

from arches.app.datatypes import datatypes
from arches.app.models import models
from django.utils.translation import get_language


class StringDataType(datatypes.StringDataType):
    def transform_value_for_tile(self, value, *, languages=None, **kwargs):
        """
        Override to:
        1. avoid refetching languages already present in kwargs.
        2. get the direction from the database when falling back to get_language().
        """
        language = get_language()
        try:
            regex = re.compile(r"(.+)\|([A-Za-z-]+)$", flags=re.DOTALL | re.MULTILINE)
            match = regex.match(value)
            if match is not None:
                language = match.groups()[1]
                value = match.groups()[0]
        except Exception as e:
            pass

        try:
            parsed_value = json.loads(value)
        except Exception:
            try:
                parsed_value = ast.literal_eval(value)
            except Exception:
                parsed_value = value

        try:
            parsed_value.keys()
            return parsed_value
        except AttributeError:
            if languages:
                for language_object in languages:
                    if language_object.code == language:
                        break
                else:
                    language_object = None
            else:
                language_object = models.Language.objects.filter(code=language).first()
            if language_object is not None:
                return {
                    language: {
                        "value": value,
                        "direction": language_object.default_direction,
                    }
                }

            return {language: {"value": value, "direction": "ltr"}}



from django.utils.translation import get_language

from arches import VERSION as arches_version
from arches.app.datatypes import datatypes
from arches.app.models import models
from arches.app.models.models import File


class FileListDataType(datatypes.FileListDataType):
    localized_metadata_keys = {"altText", "attribution", "description", "title"}

    def get_display_value(self, tile, node, **kwargs):
        data = self.get_tile_data(tile)
        files = data[str(node.nodeid)]
        file_urls = ""
        if files is not None:
            file_urls = " | ".join(
                [file["url"] or "" for file in files if "url" in file]
            )

        return file_urls

    def transform_value_for_tile(self, value, *, languages=None, **kwargs):
        if not value:
            return value
        if not languages:  # pragma: no cover
            languages = models.Language.objects.all()
        language = get_language()
        # arches == 9.0.0 - remove the stringifieid_list in favor of the 8.1.0 logic
        if arches_version < (8, 1):
            original_value = value
            if isinstance(value, str):
                stringified_list = value
            elif isinstance(value, list) and all(
                isinstance(file_info, dict) for file_info in value
            ):
                stringified_list = ",".join(
                    [file_info.get("name") for file_info in value]
                )
            else:
                raise TypeError(value)
            value = super().transform_value_for_tile(
                stringified_list, languages=languages, **kwargs
            )
            new_value = []
            for file in value:
                if not isinstance(original_value, str):
                    matching_file_info = next(
                        (
                            file_dict
                            for file_dict in original_value
                            if file_dict.get("name") == file.get("name")
                        ),
                        None,
                    )
                    if matching_file_info:
                        new_value.append({**matching_file_info, **file})
                else:
                    new_value.append(file)
        else:
            new_value = super().transform_value_for_tile(
                value, languages=languages, **kwargs
            )

        # Remove file object created in transform_value_for_tile
        # after discussion with chiatt, this behavior is only really needed
        # for the bulk loader (and causes integrity problems/duplicity) -
        # file will be recreated later in post_tile_save.  Skip this deletion
        if "bulk_import" not in kwargs and (
            "is_existing_tile" not in kwargs or not kwargs["is_existing_tile"]
        ):
            File.objects.filter(
                fileid__in=[file["file_id"] for file in new_value]
            ).delete()
            for file_dict in new_value:
                file_dict["file_id"] = None
                file_dict["url"] = None

        for file_info in new_value:
            for key, val in file_info.items():
                if key not in self.localized_metadata_keys:
                    continue
                original_val = val
                if not isinstance(original_val, dict):
                    file_info[key] = {}
                for lang in languages:
                    if lang.code not in file_info[key]:
                        file_info[key][lang.code] = {
                            "value": original_val if lang.code == language else "",
                            "direction": lang.default_direction,
                        }

        return new_value



from rest_framework.generics import (
    ListCreateAPIView,
    RetrieveAPIView,
    RetrieveUpdateDestroyAPIView,
)
from rest_framework.parsers import JSONParser

from arches_querysets.rest_framework.multipart_json_parser import MultiPartJSONParser
from arches_querysets.rest_framework.pagination import ArchesLimitOffsetPagination
from arches_querysets.rest_framework.permissions import ReadOnly, ResourceEditor
from arches_querysets.rest_framework.serializers import (
    ArchesResourceSerializer,
    ArchesResourceTopNodegroupsSerializer,
    ArchesSingleNodegroupSerializer,
    ArchesTileSerializer,
)
from arches_querysets.rest_framework.view_mixins import ArchesModelAPIMixin


class ArchesResourceListCreateView(ArchesModelAPIMixin, ListCreateAPIView):
    permission_classes = [ResourceEditor | ReadOnly]
    serializer_class = ArchesResourceSerializer
    parser_classes = [JSONParser, MultiPartJSONParser]
    pagination_class = ArchesLimitOffsetPagination


class ArchesResourceDetailView(ArchesModelAPIMixin, RetrieveUpdateDestroyAPIView):
    permission_classes = [ResourceEditor | ReadOnly]
    serializer_class = ArchesResourceSerializer
    parser_classes = [JSONParser, MultiPartJSONParser]


class ArchesTileListCreateView(ArchesModelAPIMixin, ListCreateAPIView):
    permission_classes = [ResourceEditor | ReadOnly]
    serializer_class = ArchesTileSerializer
    parser_classes = [JSONParser, MultiPartJSONParser]
    pagination_class = ArchesLimitOffsetPagination


class ArchesTileDetailView(ArchesModelAPIMixin, RetrieveUpdateDestroyAPIView):
    permission_classes = [ResourceEditor | ReadOnly]
    serializer_class = ArchesTileSerializer
    parser_classes = [JSONParser, MultiPartJSONParser]


### Views for returning blank resource & tile templates


class ArchesResourceBlankView(ArchesModelAPIMixin, RetrieveAPIView):
    permission_classes = [ReadOnly]
    serializer_class = ArchesResourceSerializer

    def get_serializer_class(self):
        if self.request.GET.get("exclude_children", "").lower() == "true":
            return ArchesResourceTopNodegroupsSerializer
        return self.serializer_class

    def get_serializer_context(self):
        serializer_context = super().get_serializer_context()
        serializer_context["fill_blanks"] = True
        return serializer_context

    def get_object(self, *args, **kwargs):
        return None


class ArchesTileBlankView(ArchesModelAPIMixin, RetrieveAPIView):
    permission_classes = [ReadOnly]
    serializer_class = ArchesTileSerializer

    def get_serializer_class(self):
        if self.request.GET.get("exclude_children", "").lower() == "true":
            return ArchesSingleNodegroupSerializer
        return self.serializer_class

    def get_serializer_context(self):
        serializer_context = super().get_serializer_context()
        serializer_context["fill_blanks"] = True
        return serializer_context

    def get_object(self, *args, **kwargs):
        return None



import copy
from uuid import uuid4
from arches.app.models.models import TileModel
from django.core.exceptions import ValidationError

from arches_querysets.models import ResourceTileTree, TileTree
from arches_querysets.utils.models import ensure_request
from arches_querysets.utils.tests import GraphTestCase


class SaveTileTests(GraphTestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        resources = ResourceTileTree.get_tiles(
            "datatype_lookups", as_representation=True
        )
        cls.resource_42 = resources.get(pk=cls.resource_42.pk)
        cls.resource_42.graph_publication_id = cls.graph.publication_id
        cls.resource_42.save()
        cls.datatype_1 = cls.resource_42.aliased_data.datatypes_1
        cls.datatype_n = cls.resource_42.aliased_data.datatypes_n

        cls.resource_none = resources.get(pk=cls.resource_none.pk)
        cls.resource_none.graph_publication_id = cls.graph.publication_id
        cls.resource_none.save()
        cls.datatype_1_none = cls.resource_none.aliased_data.datatypes_1
        cls.datatype_n_none = cls.resource_none.aliased_data.datatypes_n

    def assert_default_values_present(self, resource):
        for node_id_str, value in resource.aliased_data.datatypes_1.data.items():
            node = [node for node in self.data_nodes if str(node.pk) == node_id_str][0]
            with self.subTest(alias=node.alias):
                default_value = self.default_vals_by_nodeid[node_id_str]
                expected = TileTree.get_cleaned_default_value(node, default_value)
                self.assertEqual(value, expected)

    def test_blank_tile_save_with_defaults(self):
        # Existing tiles with `None`'s should not be updated with defaults during save
        self.resource_none.save()
        for key, value in self.resource_none.aliased_data.datatypes_1.data.items():
            self.assertIsNone(value, f"Expected None for {key}")

        # fill_blanks only intializes a tile for nodegroups that don't yet have
        # a tile. Remove those tiles so we can use fill_blanks.
        self.resource_42.aliased_data.datatypes_1.delete()
        self.resource_42.refresh_from_db()
        self.resource_42.fill_blanks()
        # Saving a blank tile should populate default values if defaults are defined.
        self.resource_42.save(force_admin=True)
        self.assert_default_values_present(self.resource_42)

        # fill_blanks gives an unsaved empty tile, but we also need to test that inserting
        # a tile (ie from the frontend) will fill defaults if no values are provided
        self.resource_42.aliased_data.datatypes_1.delete()
        self.resource_42.refresh_from_db()
        self.resource_42.fill_blanks()

        # mock a new tile via fill_blanks, but overwrite default values set by fill_blanks
        for node in self.resource_42.aliased_data.datatypes_1.data:
            self.resource_42.aliased_data.datatypes_1.data[node] = None
        # Save should stock defaults
        self.resource_42.aliased_data.datatypes_1.save(force_admin=True)
        self.assert_default_values_present(self.resource_42)

    def test_fill_blanks(self):
        self.resource_none.tilemodel_set.all().delete()
        self.resource_none.fill_blanks()
        self.assertIsInstance(self.resource_none.aliased_data.datatypes_1, TileTree)
        self.assertIsInstance(self.resource_none.aliased_data.datatypes_n[0], TileTree)
        self.assertIsInstance(
            self.resource_none.aliased_data.datatypes_1.aliased_data.datatypes_1_child,
            TileTree,
        )

        # Remove the child, fill_blanks() again.
        self.resource_none.aliased_data.datatypes_1.aliased_data.datatypes_1_child = (
            None
        )
        self.resource_none.fill_blanks()
        self.assertIsInstance(
            self.resource_none.aliased_data.datatypes_1.aliased_data.datatypes_1_child,
            TileTree,
        )

        msg = "Attempted to append to a populated cardinality-1 nodegroup"
        with self.assertRaisesMessage(RuntimeError, msg):
            self.resource_none.append_tile("datatypes_1")

    def test_parent_tile_backfilled_on_child_tile_save(self):
        self.resource_none.tilemodel_set.all().delete()
        new_child_tile = TileTree(
            resourceinstance=self.resource_none,
            nodegroup=self.nodegroup_1_child,
            number_child=4,
            # TODO(arches_version==9.0.0): in Arches 8+, data={} can be removed.
            data={},
        )
        new_child_tile.save(force_admin=True)
        # The parent property holds the richer TileTree
        self.assertIsInstance(new_child_tile.parent, TileTree)
        # The regular Django field is untouched (still a vanilla TileModel)
        self.assertNotIsInstance(new_child_tile.parenttile, TileTree)
        self.assertIsInstance(new_child_tile.parenttile, TileModel)

    def test_cardinality_error(self):
        tt = TileTree(
            nodegroup=self.nodegroup_1, resourceinstance=self.resource_42, data={}
        )
        with self.assertRaises(ValidationError) as ctx:
            tt.save(force_admin=True)
        self.assertEqual(
            ctx.exception.message_dict, {"datatypes_1": ["Tile Cardinality Error"]}
        )

    def test_simple_tile_delete(self):
        request = ensure_request(None, True)
        request.GET["delete_missing_tiles"] = "true"
        self.resource_42.aliased_data.datatypes_1 = None
        self.resource_42.save(request=request, force_admin=True)

        resources = ResourceTileTree.get_tiles(
            "datatype_lookups", as_representation=True
        )
        local_resource_42 = resources.get(pk=self.resource_42.pk)
        self.assertEqual(local_resource_42.aliased_data.datatypes_1, None)

    def test_cardinality_n_tile_delete(self):
        request = ensure_request(None, True)
        request.GET["delete_missing_tiles"] = "true"

        tile_copy = copy.deepcopy(self.datatype_n[0])
        tile_copy.tileid = uuid4()
        tile_copy.aliased_data.datatypes_n_child[0].tileid = uuid4()

        self.resource_42.aliased_data.datatypes_n.append(tile_copy)
        self.resource_42.save(request=request, force_admin=True)

        resources = ResourceTileTree.get_tiles(
            "datatype_lookups", as_representation=True
        )
        local_resource_42 = resources.get(pk=self.resource_42.pk)

        self.assertEqual(
            len(local_resource_42.aliased_data.datatypes_n),
            2,
            "Confirm two tiles before deletion",
        )

        # now test deletion by restoring the original tiles, effectively
        # removing the newly added tile and its child tile
        local_resource_42.aliased_data.datatypes_n = [self.datatype_n[0]]
        local_resource_42.save(request=request, force_admin=True)

        resources = ResourceTileTree.get_tiles(
            "datatype_lookups", as_representation=True
        )
        local_resource_42 = resources.get(pk=self.resource_42.pk)
        self.assertEqual(
            len(local_resource_42.aliased_data.datatypes_n),
            1,
            "Confirm one tile after deletion",
        )
        self.assertEqual(
            len(
                local_resource_42.aliased_data.datatypes_n[
                    0
                ].aliased_data.datatypes_n_child
            ),
            1,
            "Confirm child tile still exists after deletion of sibling tile",
        )



"""
Module to hold improved versions of datatype methods
until we can verify correctness/desirability & upstream the changes.
"""

from .concept_types import *
from .file import *
from .resource_types import *
from .node_value import *
from .string import *
from .url import *

from .datatypes import *



from arches.app.utils.betterJSONSerializer import JSONDeserializer, JSONSerializer
from arches_querysets.utils.tests import GraphTestCase


class SerializationTests(GraphTestCase):
    def test_serialization_via_better_json_serializer(self):
        dict_string = JSONSerializer().serialize(self.resource_42)
        resource_dict = JSONDeserializer().deserialize(dict_string)
        # Django model fields are present.
        self.assertIn("graph_id", resource_dict)
        tile_dict = resource_dict["aliased_data"]["datatypes_1"]
        self.assertIn("nodegroup_id", tile_dict)
        # Node values are present.
        self.assertIn("non_localized_string_alias", tile_dict["aliased_data"])
        # Special properties are not present.
        self.assertNotIn("data", tile_dict)
        self.assertNotIn("parent", tile_dict)

        # Child tiles appear under nodegroup aliases.
        child_tile_dict = tile_dict["aliased_data"]["datatypes_1_child"]
        self.assertIn("tileid", child_tile_dict)

        # Cardinality N tiles appear in an array.
        tile_list = resource_dict["aliased_data"]["datatypes_n"][0]
        child_tile_list = tile_list["aliased_data"]["datatypes_n_child"]
        self.assertIn("tileid", child_tile_list[0])



"""Wrapping the Django fields allows registering datatype-specific lookups."""

from django.contrib.postgres.fields import ArrayField
from django.db.models import JSONField


class ResourceInstanceField(JSONField):
    pass


class ResourceInstanceListField(JSONField):
    pass


class ConceptListField(JSONField):
    pass


class DomainListField(JSONField):
    pass


class LocalizedStringField(JSONField):
    pass


### Cardinality N
class CardinalityNResourceInstanceField(ArrayField):
    pass


class CardinalityNResourceInstanceListField(ArrayField):
    pass


class CardinalityNConceptListField(ArrayField):
    pass


class CardinalityNLocalizedStringField(ArrayField):
    pass


class CardinalityNTextField(ArrayField):
    pass


class CardinalityNUUIDField(ArrayField):
    pass


class CardinalityNJSONField(ArrayField):
    pass



from django.db import models

from arches import VERSION as arches_version
from arches.app.models.models import Node


def get_nodegroup_alias_lookup(graph_slug):
    filters = models.Q(pk=models.F("nodegroup_id"), graph__slug=graph_slug)
    # arches_version==9.0.0
    if arches_version >= (8, 0):
        filters &= models.Q(source_identifier=None)
    return {node.pk: node.alias for node in Node.objects.filter(filters).only("alias")}



import logging
import uuid
from itertools import chain

from arches import VERSION as arches_version
from arches.app.datatypes import datatypes
from arches.app.models import models
from django.utils.translation import get_language
from django.utils.translation import gettext as _

logger = logging.getLogger(__name__)


class ResourceInstanceDataType(datatypes.ResourceInstanceDataType):
    def transform_value_for_tile(self, value, **kwargs):
        parsed = super().transform_value_for_tile(value, **kwargs)
        if parsed is None:
            parsed = value

        graph_configs_by_graph_id = {
            graph_config["graphid"]: graph_config
            for graph_config in kwargs.get("graphs", [])
        }

        if not isinstance(parsed, list):
            parsed = [parsed]
        transformed = []
        for inner_val in parsed:
            match inner_val:
                case models.ResourceInstance():
                    transformed.append(
                        self.from_id_string(
                            str(inner_val.pk),
                            graph_configs_by_graph_id.get(inner_val.graph_id),
                        )
                    )
                case uuid.UUID():
                    # TODO: handle multiple graph configs, requires db?
                    transformed.append(self.from_id_string(str(inner_val)))
                case str():
                    # TODO: handle multiple graph configs, requires db?
                    transformed.append(self.from_id_string(inner_val))
                case _:
                    transformed.append(inner_val)
        return transformed

    def get_resource(self, tile):
        try:
            return tile.resourceinstance
        except models.ResourceInstance.DoesNotExist:
            return None
        except AttributeError:
            resource_id = tile["resourceinstance_id"]
            return models.ResourceInstance.objects.filter(pk=resource_id).first()

    def get_display_value(self, tile, node, *, details=None, **kwargs):
        if details is None:
            data = self.get_tile_data(tile)
            value = data.get(str(node.nodeid))
            resource = self.get_resource(tile)
            details = self.get_details(value, resource=resource)
        return ", ".join(
            [detail["display_value"] or "" for detail in details if detail]
        )

    def to_python(self, value, *, resource=None):
        if not (related_resources := self.get_related_resources(value, resource)):
            return None
        return related_resources[0]

    def get_related_resources(self, value, resource):
        if not value:
            return []
        related_resources = []

        # arches_version==9.0.0
        if arches_version >= (8, 0):
            relations = resource.from_resxres.all()
        else:
            relations = resource.resxres_resource_instance_ids_from.all()

        def handle_missing_data(to_resource_id):
            msg = f"Missing ResourceXResource target: {to_resource_id}"
            logger.warning(msg)

        for inner_val in value:
            if not inner_val:
                continue
            target_resource_id = uuid.UUID(inner_val["resourceId"])
            if not relations:
                # arches_version==9.0.0
                if arches_version >= (8, 0):
                    relations = models.ResourceXResource.objects.filter(
                        to_resource_id=target_resource_id
                    ).select_related("to_resource")
                else:
                    relations = models.ResourceXResource.objects.filter(
                        resourceinstanceidto_id=target_resource_id
                    ).select_related("resourceinstanceidto")

            for relation in relations:
                to_resource_id = (
                    relation.resourceinstanceidto_id
                    # arches_version==9.0.0
                    if arches_version < (8, 0)
                    else relation.to_resource_id
                )
                if to_resource_id == target_resource_id:
                    try:
                        to_resource = (
                            relation.resourceinstanceidto
                            # arches_version==9.0.0
                            if arches_version < (8, 0)
                            else relation.to_resource
                        )
                        if to_resource is None:
                            raise models.ResourceInstance.DoesNotExist
                    except models.ResourceInstance.DoesNotExist:
                        handle_missing_data(to_resource_id)
                        break
                    related_resources.append(to_resource)
                    break

        return related_resources

    def get_details(self, value, *, resource=None, **kwargs):
        lang = get_language()
        related_resources_by_id = {
            related_resource.pk: related_resource
            for related_resource in self.get_related_resources(value, resource)
        }
        ret = []
        for inner_val in value or []:
            if not inner_val:
                continue
            if related := related_resources_by_id.get(
                uuid.UUID(inner_val["resourceId"]), None
            ):
                ret.append(
                    {
                        "resource_id": str(related.pk),
                        # TODO: gracefully handle missing language.
                        "display_value": related.descriptors[lang]["name"],
                    }
                )
            else:
                ret.append(
                    {
                        "resource_id": None,
                        "display_value": _("Missing"),
                    }
                )
        return ret

    @staticmethod
    def from_id_string(uuid_string, graph_config=None):
        if graph_config is None:
            graph_config = {}
        return {
            "resourceId": uuid_string,
            "ontologyProperty": graph_config.get("ontologyProperty", ""),
            "inverseOntologyProperty": graph_config.get("inverseOntologyProperty", ""),
        }

    @staticmethod
    def values_match(value1, value2):
        if not isinstance(value1, list) or not isinstance(value2, list):
            return value1 == value2
        copy1 = [{**inner_val} for inner_val in value1]
        copy2 = [{**inner_val} for inner_val in value2]
        for inner_val in chain(copy1, copy2):
            inner_val.pop("resourceXresourceId", None)
        return copy1 == copy2


class ResourceInstanceListDataType(ResourceInstanceDataType):
    def collects_multiple_values(self):
        return True

    def to_python(self, value, *, resource=None):
        if not (related_resources := self.get_related_resources(value, resource)):
            return None
        return related_resources

    def get_details(self, value, *, resource=None, **kwargs):
        if not value:
            return []
        details = super().get_details(value=value, resource=resource, **kwargs)
        resource_display_value_map = {
            str(detail["resource_id"]): detail["display_value"] for detail in details
        }
        return [
            {
                "resource_id": resource_dict["resourceId"],
                "display_value": resource_display_value_map.get(
                    resource_dict["resourceId"], _("Missing")
                ),
            }
            for resource_dict in value
        ]



from arches_querysets.bulk_operations.tiles import TileTreeOperation
from arches_querysets.models import TileTree
from arches_querysets.utils.models import ensure_request
from arches_querysets.utils.tests import GraphTestCase


class TileTreeOperationTests(GraphTestCase):
    def test_partial_update_leaves_missing_nodes_alone(self):
        request = ensure_request(None, force_admin=True)
        test_tiles = TileTree.get_tiles("datatype_lookups", "datatypes_1")
        entry_tile = test_tiles.get(pk=self.cardinality_1_tile.pk)
        del entry_tile.aliased_data.boolean_alias
        operation = TileTreeOperation(entry=entry_tile, request=request, partial=True)
        operation.validate()
        with self.assertRaises(AttributeError):
            entry_tile.aliased_data.boolean_alias

    def test_full_update_resets_missing_nodes_to_default(self):
        request = ensure_request(None, force_admin=True)
        test_tiles = TileTree.get_tiles("datatype_lookups", "datatypes_1")
        entry_tile = test_tiles.get(pk=self.cardinality_1_tile.pk)
        self.assertIs(entry_tile.aliased_data.boolean_alias, True)
        del entry_tile.aliased_data.boolean_alias
        operation = TileTreeOperation(entry=entry_tile, request=request, partial=False)
        operation.validate()
        self.assertIs(entry_tile.aliased_data.boolean_alias, False)



import logging
import uuid
from types import SimpleNamespace
from typing import Mapping

from django.conf import settings
from django.core.exceptions import (
    MultipleObjectsReturned,
    ObjectDoesNotExist,
    ValidationError,
)
from django.db import models
from django.utils.translation import gettext as _

from arches import VERSION as arches_version
from arches.app.models.models import GraphModel, ResourceInstance, TileModel
from arches.app.models.resource import Resource
from arches.app.models.tile import Tile
from arches.app.utils.betterJSONSerializer import JSONSerializer
from arches.app.utils.permission_backend import user_is_resource_reviewer

from arches_querysets.bulk_operations.tiles import TileTreeOperation
from arches_querysets.datatypes.datatypes import DataTypeFactory
from arches_querysets.lookups import *  # registers lookups
from arches_querysets.querysets import (
    GraphWithPrefetchingQuerySet,
    ResourceTileTreeQuerySet,
    TileTreeManager,
    TileTreeQuerySet,
)
from arches_querysets.utils.models import (
    append_tiles_recursively,
    ensure_request,
    pop_arches_model_kwargs,
)
from arches_querysets.tasks import index_resource
import arches.app.utils.task_management as task_management

logger = logging.getLogger(__name__)


class AliasedData(SimpleNamespace):
    """Provides dot access into node values and nested nodegroups by alias.

    >>> ResourceTileTree.get_tiles('new_resource_model_1').get(...).aliased_data
    AliasedData(string_node={'en': {'value': 'abcde', 'direction': 'ltr'}},
                child_node=<TileTree: child_node (c3637412-9b13-4f05-8f4a-5a80560b8b6e)>)
    """

    def serialize(self, **kwargs):
        serializer = JSONSerializer()
        serializer.force_recalculation = kwargs.get("force_recalculation", False)
        return {
            key: serializer.handle_object(val, **kwargs)
            for key, val in vars(self).items()
        }

    def __iter__(self):
        return iter(self._keys())

    def _keys(self):
        return vars(self).keys()

    def _values(self):
        return vars(self).values()

    def _items(self):
        return vars(self).items()


class AliasedDataMixin:
    """Don't implement properties: https://github.com/archesproject/arches/issues/12310"""

    def _refresh_aliased_data(self, using, fields, from_queryset):
        try:
            del self._tile_trees
        except AttributeError:
            pass

        # Commandeer the responsibility for filtering on pk from Django
        # so we can retrieve aliased data from the queryset cache.
        from_queryset = from_queryset.filter(pk=self.pk)
        # arches_version==9.0.0
        if arches_version >= (8, 0):
            # Patch out filter(pk=...) so that when refresh_from_db() calls get(),
            # it populates the cache. TODO: ask on forum about happier path.
            from_queryset.filter = lambda pk=None: from_queryset
            models.Model.refresh_from_db(self, using, fields, from_queryset)
            # Retrieve aliased data from the queryset cache.
            self.aliased_data = from_queryset[0].aliased_data
            self._tile_trees = from_queryset[0]._tile_trees
        else:
            # Django 4: good-enough riff on refresh_from_db(), but not bulletproof.
            db_instance = from_queryset.get()
            for field in db_instance._meta.concrete_fields:
                setattr(self, field.attname, getattr(db_instance, field.attname))
            self.aliased_data = db_instance.aliased_data
            if isinstance(self, TileModel) and self.parenttile_id:
                self.parenttile = TileModel.objects.get(pk=self.parenttile_id)
            self._tile_trees = from_queryset[0]._tile_trees


class ResourceTileTree(ResourceInstance, AliasedDataMixin):
    objects = ResourceTileTreeQuerySet.as_manager()

    class Meta:
        proxy = True
        db_table = "resource_instances"
        permissions = (("no_access_to_resourceinstance", "No Access"),)

    def __init__(self, *args, **kwargs):
        self._as_representation = kwargs.pop("__as_representation", False)
        self._request = kwargs.pop("__request", None)
        arches_model_kwargs, other_kwargs = pop_arches_model_kwargs(
            kwargs, self._meta.get_fields()
        )
        super().__init__(*args, **other_kwargs)
        self.aliased_data = arches_model_kwargs.pop(
            "aliased_data", None
        ) or AliasedData(**arches_model_kwargs)
        self._sealed = False

    @property
    def aliased_data(self):
        return self._aliased_data

    @aliased_data.setter
    def aliased_data(self, value):
        self._aliased_data = value

    @property
    def sealed(self):
        """
        A boolean set when this instance is yielded from a QuerySet signifying its
        related .graph has been replaced with an instance w/ related objects prefetched.
        """
        return self._sealed

    @sealed.setter
    def sealed(self, value):
        self._sealed = value

    def save(
        self, *, request=None, index=True, partial=None, force_admin=False, **kwargs
    ):
        """
        If `partial` is not explicitly provided, infer it from the HTTP method:

        - PATCH -> partial=True  (diff-only)
        - PUT   -> partial=False (full replace; delete/reset missing)
        - POST  -> partial=False (new resource; full source of truth)
        - anything else with no request -> partial=True as a safe default
        """
        # arches_version==9.0.0
        if (
            arches_version >= (8, 0)
            and self.graph_publication_id
            and (self.graph_publication_id != self.graph.publication_id)
        ):
            raise ValidationError(_("Graph Has Different Publication"))

        request = ensure_request(request or self._request, force_admin)

        if partial is None:
            request_method_value = getattr(request, "method", "") if request else ""
            http_method = str(request_method_value or "").upper()

            if http_method == "PATCH":
                partial = True
            elif http_method in {"PUT", "POST"}:
                partial = False
            else:
                partial = True

        self._save_aliased_data(
            request=request,
            index=index,
            partial=partial,
            force_admin=force_admin,
            **kwargs,
        )

    @classmethod
    def get_tiles(
        cls,
        graph_slug,
        *,
        resource_ids=None,
        as_representation=False,
        nodes=None,
    ):
        """Return a chainable QuerySet for a requested graph's instances,
        with tile data keyed by node and nodegroup aliases.

        See `arches_querysets.querysets.ResourceTileTreeQuerySet.get_tiles`.
        """
        return cls.objects.get_tiles(
            graph_slug,
            resource_ids=resource_ids,
            as_representation=as_representation,
            nodes=nodes,
        )

    def append_tile(self, nodegroup_alias):
        grouping_node_aliases = {
            node.alias
            for node in self.graph.node_set.all()
            if node.pk == node.nodegroup_id
        }
        if nodegroup_alias not in grouping_node_aliases:
            raise ValueError(nodegroup_alias)

        TileTree.create_blank_tile(
            nodegroup_alias=nodegroup_alias,
            container=self,
            graph_nodes=self.graph.node_set.all(),
        )

    def fill_blanks(self):
        """Initialize a blank tile with empty values for each nodegroup lacking a tile."""
        append_tiles_recursively(self)

    def save_edit(self, user=None, transaction_id=None, *, edit_type=None):
        """Intended to replace proxy model method eventually."""
        if self._state.adding:
            edit_type = "create"
        if not edit_type:
            return

        # Until save_edit() is a static method, work around it.
        ephemeral_proxy_instance = Resource()
        ephemeral_proxy_instance.graphid = self.graph_id
        ephemeral_proxy_instance.resourceinstanceid = str(self.pk)
        ephemeral_proxy_instance.save_edit(
            user=user, edit_type=edit_type, transaction_id=transaction_id
        )

    def refresh_from_db(self, using=None, fields=None, from_queryset=None):
        if from_queryset is None:
            # TODO: symptom that we need a backreference to the queryset args.
            from_queryset = self.__class__.get_tiles(
                self.graph.slug,
                as_representation=getattr(self, "_as_representation", False),
            )
        self._refresh_aliased_data(using, fields, from_queryset)

    def _save_aliased_data(
        self, *, request=None, index=True, partial=True, force_admin=False, **kwargs
    ):
        """Raises a compound ValidationError with any failing tile values."""
        request = ensure_request(request, force_admin)
        operation = TileTreeOperation(
            entry=self, request=request, partial=partial, save_kwargs=kwargs
        )
        for_new_resource = self._state.adding

        # This will also call ResourceInstance.save()
        operation.validate_and_save_tiles()
        if not self.pk:
            self.pk = operation.resourceid

        # Run side effects trapped on Resource.save()
        proxy_resource = (
            Resource.objects.filter(pk=self.pk)
            .select_related("graph__publication")
            .get()
        )
        proxy_resource.save_descriptors()
        if index:
            if (
                getattr(settings, "EVENTUALLY_CONSISTENT_ES_INDEXING", False)
                and task_management.check_if_celery_available()
            ):
                index_resource.apply_async((self.resourceinstanceid,))
            else:
                proxy_resource.index()

        # arches_version==9.0.0
        if arches_version < (8, 0) and request and for_new_resource:
            self.save_edit(
                user=request.user,
                transaction_id=operation.transaction_id,
                edit_type="create",
            )

        self.refresh_from_db(
            using=kwargs.get("using"), fields=kwargs.get("update_fields")
        )
        if request.GET.get("fill_blanks", "f").lower().startswith("t"):
            self.fill_blanks()


class TileTree(TileModel, AliasedDataMixin):
    objects = TileTreeManager.from_queryset(TileTreeQuerySet)()

    class Meta:
        proxy = True
        db_table = "tiles"

    def __init__(self, *args, **kwargs):
        self._as_representation = kwargs.pop("__as_representation", False)
        self._request = kwargs.pop("__request", None)
        arches_model_kwargs, other_kwargs = pop_arches_model_kwargs(
            kwargs, self._meta.get_fields()
        )
        super().__init__(*args, **other_kwargs)
        self.aliased_data = arches_model_kwargs.pop(
            "aliased_data", None
        ) or AliasedData(**arches_model_kwargs)
        self._parent = None
        self._sealed = False

    @property
    def aliased_data(self):
        return self._aliased_data

    @aliased_data.setter
    def aliased_data(self, value):
        self._aliased_data = value

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent):
        self._parent = parent

    @property
    def sealed(self):
        """
        A boolean set when this instance is yielded from a QuerySet signifying its
        related .nodegroup and .resourceinstance.graph have been replaced with
        instances w/ related objects prefetched.
        """
        return self._sealed

    @sealed.setter
    def sealed(self, value):
        self._sealed = value

    def save(
        self, *, request=None, index=True, partial=True, force_admin=False, **kwargs
    ):
        """
        partial=True (HTTP PATCH): absent nodes ignored, absent child tiles ignored.
        partial=False (HTTP PUT): absent nodes reset to default, absent child tiles deleted.
        """
        # arches_version==9.0.0
        if (
            arches_version >= (8, 0)
            and self.resourceinstance_id
            and self.resourceinstance.graph_publication_id
            and (
                self.resourceinstance.graph_publication_id
                != self.resourceinstance.graph.publication_id
            )
        ):
            raise ValidationError(_("Graph Has Different Publication"))

        request = request or self._request
        # Mimic some computations trapped on TileModel.save().
        # arches_version==9.0.0
        if (
            arches_version >= (8, 0)
            and self.sortorder is None
            or self.is_fully_provisional()
        ):
            self.set_next_sort_order()
        self._save_aliased_data(
            request=request,
            index=index,
            partial=partial,
            force_admin=force_admin,
            **kwargs,
        )

    @classmethod
    def get_tiles(
        cls,
        graph_slug,
        nodegroup_alias,
        *,
        resource_ids=None,
        as_representation=False,
        nodes=None,
        depth=20,
    ):
        """See `arches_querysets.querysets.TileTreeQuerySet.get_tiles`."""
        return cls.objects.get_tiles(
            graph_slug=graph_slug,
            nodegroup_alias=nodegroup_alias,
            resource_ids=resource_ids,
            as_representation=as_representation,
            nodes=nodes,
            depth=depth,
        )

    def serialize(self, **kwargs):
        """Prevent serialization of the vanilla .data field, as well as properties
        (serializing .parent would cause cycles)."""
        options = {**kwargs}
        ignored_props = {"data", "parent", "sealed"}
        options["exclude"] = ignored_props | set(options.pop("exclude", {}))
        return JSONSerializer().handle_model(self, **options)

    def clean_fields(self, exclude=None):
        if (
            self.nodegroup
            and self.nodegroup.parentnodegroup_id
            and "parenttile" not in exclude
        ):
            if (
                not self.parenttile_id
                or self.nodegroup.parentnodegroup_id != self.parenttile.nodegroup_id
            ):
                msg = _("Wrong parent tile for parent nodegroup.")
                raise ValidationError({self.find_nodegroup_alias(): msg})
        # Exclude parenttile to ensure batch creations of parent & child do not fail.
        new_exclude = [*(exclude or []), "parenttile"]
        super().clean_fields(exclude=new_exclude)

    def find_nodegroup_alias(self, grouping_node_lookup=None):
        # arches_version==9.0.0
        if arches_version >= (8, 0):
            if grouping_node_lookup:
                self.nodegroup.grouping_node = grouping_node_lookup[self.nodegroup_id]
            return super().find_nodegroup_alias()
        if not getattr(self, "_nodegroup_alias", None):
            # TileTreeManager provides "_nodegroup_alias" annotation on 7.6, but perform
            # a last-minute check just in case. Also runs if the node alias is null.
            for node in self.nodegroup.node_set.all():
                if node.pk == self.nodegroup.pk:
                    self._nodegroup_alias = node.alias
        return self._nodegroup_alias

    @classmethod
    def deserialize(cls, tile_dict, parent_tile: TileModel | None):
        """
        If you're not using the Django REST Framework optional dependency,
        e.g. if you evaluate a queryset with as_representation=True and resave
        the instance, you'll need a way to deserialize dicts into TileTrees.
        """
        if not isinstance(tile_dict, Mapping):
            raise TypeError(
                f'Expected a mapping, got: "{tile_dict}". '
                "Did you mistakenly provide node data directly under a nodegroup alias?"
            )
        attrs = {**tile_dict}
        if (tile_id := attrs.pop("tileid", None)) and isinstance(tile_id, str):
            attrs["tileid"] = uuid.UUID(tile_id)
        if (resourceinstance_id := attrs.pop("resourceinstance", None)) and isinstance(
            resourceinstance_id, str
        ):
            attrs["resourceinstance_id"] = uuid.UUID(resourceinstance_id)
        if (nodegroup_id := attrs.pop("nodegroup", None)) and isinstance(
            nodegroup_id, str
        ):
            attrs["nodegroup_id"] = uuid.UUID(nodegroup_id)
        if (parenttile_id := attrs.pop("parenttile", None)) and isinstance(
            parenttile_id, str
        ):
            attrs["parenttile_id"] = uuid.UUID(parenttile_id)

        attrs["parenttile"] = parent_tile

        tile = cls(**attrs)
        for attr in {"resourceinstance", "nodegroup", "parenttile"}:
            if attr in tile_dict:
                try:
                    tile_dict[attr] = getattr(tile, attr)
                except:
                    pass

        # arches_version==9.0.0
        if arches_version < (8, 0):
            # Simulate the default supplied by v8.
            tile.data = {}

        return tile

    def sync_private_attributes(self, source):
        if isinstance(source, models.QuerySet):
            self._as_representation = source._hints.get("as_representation", False)
        else:
            self._as_representation = source._as_representation

    def append_tile(self, nodegroup_alias):
        grouping_node_aliases = {
            node.alias
            for node in self.resourceinstance.graph.node_set.all()
            if node.pk == node.nodegroup_id
        }
        if nodegroup_alias not in grouping_node_aliases:
            raise ValueError(nodegroup_alias)
        TileTree.create_blank_tile(
            nodegroup_alias=nodegroup_alias,
            container=self,
            graph_nodes=(
                self.resourceinstance.graph.node_set.all()
                if self.resourceinstance_id
                else None
            ),
        )

    @classmethod
    def create_blank_tile(
        cls,
        *,
        nodegroup=None,
        nodegroup_alias=None,
        container,
        graph_nodes,
        caller=None,
    ):
        """
        Provide either a nodegroup model instance or a nodegroup_alias for which
        to create a blank tile populated with default node values. If `container`
        is a ResourceTileTree or TileTree, the new tile will be inserted into the
        container's aliased_data, and blank child tiles will be created as well.
        If `container` is None, the operation is done in the reverse direction:
        the new tile becomes the container (parent), and `caller` is inserted into
        the new tile's aliased_data.
        """
        if not nodegroup and not nodegroup_alias:
            raise ValueError("nodegroup or nodegroup_alias is required.")
        nodegroup, nodegroup_alias = cls.find_nodegroup_and_alias_from_alias_or_pk(
            nodegroup_alias,
            pk=nodegroup.pk if nodegroup else None,
            graph_nodes=graph_nodes,
        )

        parent_tile = None
        if container is None:
            if not isinstance(caller, TileTree):
                raise ValueError
            resource = caller.resourceinstance
        elif isinstance(container, ResourceInstance):
            resource = container
            caller = container
        elif isinstance(container, TileModel):
            resource = container.resourceinstance
            parent_tile = container
            caller = container
        else:  # pragma: no cover
            raise ValueError

        blank_tile = cls(
            resourceinstance=resource,
            nodegroup=nodegroup,
            parenttile=parent_tile,
            data={
                str(node.pk): cls.get_default_value(node)
                for node in nodegroup.node_set.all()
                if node.datatype != "semantic"
            },
        )
        blank_tile.sync_private_attributes(caller)

        # Finalize the aliased data according to the value of self._as_representation.
        # (e.g. either a dict of node_value, display_value, & details, or call to_python().)
        for node in nodegroup.node_set.all():
            if node.datatype != "semantic":
                node_value = blank_tile.data.get(str(node.pk))
                blank_tile.set_aliased_data(node, node_value)

        def insert_into_aliased_data(nodegroup_alias, item, target):
            try:
                aliased_data_value = getattr(target.aliased_data, nodegroup_alias)
            except AttributeError:
                aliased_data_value = None if nodegroup.cardinality == "1" else []
                setattr(target.aliased_data, nodegroup_alias, aliased_data_value)
            if isinstance(aliased_data_value, list):
                aliased_data_value.append(item)
            elif aliased_data_value is None:
                setattr(target.aliased_data, nodegroup_alias, item)
            else:
                msg = "Attempted to append to a populated cardinality-1 nodegroup"
                raise RuntimeError(msg)

        if container is None:
            insert_into_aliased_data(
                caller.find_nodegroup_alias(), item=caller, target=blank_tile
            )
        else:
            insert_into_aliased_data(nodegroup_alias, item=blank_tile, target=container)

            children = (
                nodegroup.children.all()
                # arches_version==9.0.0
                if arches_version >= (8, 0)
                else nodegroup.nodegroup_set.all()
            )
            for child_nodegroup in children:
                cls.create_blank_tile(
                    nodegroup=child_nodegroup,
                    container=blank_tile,
                    graph_nodes=graph_nodes,
                )

        return blank_tile

    def fill_blanks(self):
        """Initialize a blank tile with empty values for each nodegroup lacking a tile."""
        append_tiles_recursively(self)

    @staticmethod
    def find_nodegroup_and_alias_from_alias_or_pk(alias=None, *, graph_nodes, pk=None):
        for node in graph_nodes:
            # arches_version==9.0.0: `if alias` is redundant, alias is not nullable in v8
            if (alias and node.alias == alias) or node.pk == pk:
                return (node.nodegroup, node.alias)
        raise RuntimeError

    @staticmethod
    def get_default_value(node):
        datatype_factory = DataTypeFactory()
        # TODO: When ingesting this into core, make this a method on the node.
        try:
            widget_config = node.cardxnodexwidget_set.all()[0].config
            localized_config = widget_config.serialize()
        except (IndexError, ObjectDoesNotExist, MultipleObjectsReturned):
            d_data_type = datatype_factory.datatypes[node.datatype]
            default_widget = d_data_type.defaultwidget
            localized_config = default_widget.defaultconfig
        default_value = localized_config.get("defaultValue")
        return TileTree.get_cleaned_default_value(node, default_value)

    @staticmethod
    def get_cleaned_default_value(node, default_value):
        """
        Empty strings can break type coercion at the DRF layer, e.g.
        float(""), or datatype methods that expect UUID | None.
        There should probably be some validation in the datatype
        methods to ensure that poor types don't end up in defaultValue.
        https://github.com/archesproject/arches/issues/8715#issuecomment-3033192406
        """
        dt_instance = DataTypeFactory().get_instance(node.datatype)
        node_id_str = str(node.pk)
        mock_tile = SimpleNamespace(data={node_id_str: default_value})
        dt_instance.clean(mock_tile, node_id_str)
        cleaned_default = mock_tile.data[node_id_str]

        return cleaned_default

    def get_value_with_context(self, node, node_value, datatype_contexts=None):
        datatype_instance = DataTypeFactory().get_instance(node.datatype)
        empty_display_values = (None, "", '{"url": "", "url_label": ""}')
        compiled_json = datatype_instance.to_json(self, node)
        if datatype_contexts is None:
            datatype_contexts = {}
        ret = {
            "node_value": node_value,
            "display_value": compiled_json["@display_value"],
            "details": datatype_instance.get_details(
                node_value,
                datatype_context=datatype_contexts.get(node.datatype),
                # An optional extra hint for the ResourceInstance{list} types
                # so that prefetched related resources can be used.
                resource=self.resourceinstance if self.resourceinstance_id else None,
            ),
        }
        if ret["details"] is None:
            ret["details"] = []
        if ret["display_value"] in empty_display_values:
            # Future: upstream this into datatype methods (another hook?)
            ret["display_value"] = ""
        return ret

    def set_aliased_data(self, node, node_value, datatype_contexts=None):
        """Format node_value according to the self._as_representation flag and
        set it on self.aliased_data."""
        datatype_instance = DataTypeFactory().get_instance(node.datatype)

        if self._as_representation:
            final_val = self.get_value_with_context(
                node, node_value, datatype_contexts=datatype_contexts
            )
        else:
            if hasattr(datatype_instance, "to_python"):
                resource = self.resourceinstance if self.resourceinstance_id else None
                final_val = datatype_instance.to_python(node_value, resource=resource)
            else:
                final_val = node_value

        setattr(self.aliased_data, node.alias, final_val)

    def _save_aliased_data(
        self, *, request=None, index=True, partial=True, force_admin=False, **kwargs
    ):
        request = ensure_request(request, force_admin)
        # The `entry` for the TileTreeOperation is usually self, but it could
        # be a higher parent tile in the tree if blank parents were backfilled.
        entry = self.backfill_parent_tiles()
        operation = TileTreeOperation(
            entry=entry, request=request, partial=partial, save_kwargs=kwargs
        )
        operation.validate_and_save_tiles()

        proxy_resource = (
            Resource.objects.filter(pk=self.resourceinstance_id)
            .select_related("graph__publication")
            .get()
        )
        proxy_resource.save_descriptors()

        if index:
            if (
                getattr(settings, "EVENTUALLY_CONSISTENT_ES_INDEXING", False)
                and task_management.check_if_celery_available()
            ):
                index_resource.apply_async((self.resourceinstanceid,))
            else:
                proxy_resource.index()

        self.refresh_from_db(
            using=kwargs.get("using", None),
            fields=kwargs.get("update_fields", None),
        )

    def refresh_from_db(self, using=None, fields=None, from_queryset=None):
        if from_queryset is None:
            # TODO: symptom that we need a backreference to the queryset args.
            from_queryset = self.__class__.get_tiles(
                self.resourceinstance.graph.slug,
                nodegroup_alias=self.find_nodegroup_alias(),
                as_representation=getattr(self, "_as_representation", False),
            )
        self._refresh_aliased_data(using, fields, from_queryset)

    def backfill_parent_tiles(self):
        if self.nodegroup.parentnodegroup_id and not self.parenttile_id:
            self.parenttile = self.parent = TileTree.create_blank_tile(
                nodegroup=self.nodegroup.parentnodegroup,
                container=None,
                graph_nodes=self.resourceinstance.graph.node_set.all(),
                caller=self,
            )
            return self.parenttile.backfill_parent_tiles()
        return self

    def _tile_update_is_noop(self, original_data):
        """Skipping no-op tile saves avoids regenerating RxR rows, at least
        given the current implementation that doesn't serialize them."""

        datatype_factory = DataTypeFactory()
        for node in self.nodegroup.node_set.all():
            if node.datatype == "semantic":
                continue
            node_id_str = str(node.nodeid)
            old = original_data.get(node_id_str)
            datatype_instance = datatype_factory.get_instance(node.datatype)
            new = self.data[node_id_str]
            if not datatype_instance.values_match(old, new):
                return False

        return True

    def _apply_provisional_edit(
        self, proxy, existing_data, existing_provisional_edits, *, user=None
    ):
        # TODO: decompose this out of Tile.save() and call *that*.
        # this section moves the data over from self.data to self.provisionaledits if certain users permissions are in force
        # then self.data is restored from the previously saved tile data

        oldprovisionalvalue = None
        newprovisionalvalue = None
        provisional_edit_log_details = None
        creating_new_tile = self._state.adding
        existing_instance = Tile(data={**existing_data} if existing_data else None)
        existing_instance.provisional_edits = (
            {**existing_provisional_edits} if existing_provisional_edits else None
        )
        existing_instance._state.adding = creating_new_tile
        if user is not None and not user_is_resource_reviewer(user):
            if creating_new_tile:
                # the user has previously edited this tile
                proxy.apply_provisional_edit(
                    user, self.data, action="update", existing_model=existing_instance
                )
                oldprovisional = proxy.get_provisional_edit(existing_instance, user)
                if oldprovisional is not None:
                    oldprovisionalvalue = oldprovisional["value"]
            else:
                proxy.apply_provisional_edit(user, data=self.data, action="create")

            newprovisionalvalue = self.data
            self.provisionaledits = proxy.provisionaledits
            self.data = existing_data
            # Also update proxy, which will be used to run further side effects.
            proxy.provisionaledits = proxy.provisionaledits
            proxy.data = existing_data

            provisional_edit_log_details = {
                "user": user,
                "provisional_editor": user,
                "action": "create tile" if creating_new_tile else "add edit",
            }

        return oldprovisionalvalue, newprovisionalvalue, provisional_edit_log_details


class GraphWithPrefetching(GraphModel):
    objects = GraphWithPrefetchingQuerySet.as_manager()

    class Meta:
        proxy = True
        db_table = "graphs"



from itertools import chain

from django.contrib.auth.models import User
from django.contrib.postgres.expressions import ArraySubquery
from django.contrib.postgres.fields import ArrayField
from django.db.models import (
    Case,
    ExpressionWrapper,
    F,
    JSONField,
    OuterRef,
    TextField,
    UUIDField,
    Value,
    When,
)
from django.db.models.functions import Cast, NullIf
from django.db.models.fields.json import KT
from django.http import HttpRequest
from django.utils.functional import cached_property

from arches import VERSION as arches_version
from arches.app.models.models import ResourceInstance, TileModel

from arches_querysets.datatypes.datatypes import DataTypeFactory
from arches_querysets.fields import (
    CardinalityNConceptListField,
    CardinalityNJSONField,
    CardinalityNResourceInstanceField,
    CardinalityNResourceInstanceListField,
    CardinalityNLocalizedStringField,
    CardinalityNTextField,
    CardinalityNUUIDField,
    ConceptListField,
    LocalizedStringField,
    ResourceInstanceField,
    ResourceInstanceListField,
)


DATATYPES_NEEDING_KEY_TEXT_TRANSFORM = {
    "non-localized-string",
    "date",
    "concept",
    "concept-list",
    "node-value",
}
DATATYPES_NEEDING_CAST = {"boolean", "concept-list", "date", "number"}


class CardinalityNSubquery(ArraySubquery):
    @cached_property
    def output_field(self):
        match self.query.output_field:
            case ResourceInstanceField():
                array_wrapper = CardinalityNResourceInstanceField
            case ResourceInstanceListField():
                array_wrapper = CardinalityNResourceInstanceListField
            case LocalizedStringField():
                array_wrapper = CardinalityNLocalizedStringField
            case UUIDField():
                array_wrapper = CardinalityNUUIDField
            case ConceptListField():
                array_wrapper = CardinalityNConceptListField
            case JSONField():  # e.g. url
                array_wrapper = CardinalityNJSONField
            case TextField():
                array_wrapper = CardinalityNTextField
            case _:
                array_wrapper = ArrayField
        return array_wrapper(self.query.output_field)


def field_attnames(instance_or_class):
    return {f.attname for f in instance_or_class._meta.fields}


def get_invalid_aliases(instance_or_class):
    # From QuerySet._annotate() in Django
    return set(
        chain.from_iterable(
            (
                (field.name, field.attname)
                if hasattr(field, "attname")
                else (field.name,)
            )
            for field in instance_or_class._meta.get_fields()
        )
    )


def generate_node_alias_expressions(model, nodes):
    alias_expressions = {}
    invalid_names = get_invalid_aliases(model)

    for node in nodes:
        alias = node.alias
        if node.alias in invalid_names:
            # TODO (Arches 8.1): determine reserved namespace for node aliases
            alias = "_arches_querysets_" + alias

        if issubclass(model, ResourceInstance):
            tile_values_query = get_tile_values_for_resource(node, nodes)
        elif issubclass(model, TileModel):
            tile_values_query = get_node_value_expression(node, False)
        else:
            raise ValueError
        alias_expressions[alias] = tile_values_query

    if not alias_expressions:
        raise ValueError("All fields were excluded.")

    return alias_expressions


def pop_arches_model_kwargs(kwargs, model_fields):
    arches_model_data = {}
    # Combine these sets to get both "nodegroup" and "nodegroup_id"
    model_field_names = {f.name for f in model_fields} | {
        getattr(f, "attname", None) for f in model_fields
    }
    for kwarg, value in kwargs.items():
        if kwarg not in model_field_names:
            arches_model_data[kwarg] = value
    without_model_data = {k: v for k, v in kwargs.items() if k not in arches_model_data}
    return arches_model_data, without_model_data


def get_tile_values_for_resource(node, graph_nodes):
    """
    Return a tile values query expression for use in a ResourceTileTreeQuerySet.

    Allows shallow filtering, e.g. concepts.filter(uri_content...
    even if `uri_content` is not in the top nodegroup. For this reason,
    multiple tiles for cardinality-1 nodegroups might appear if there
    are cardinality-N parents anywhere.
    """
    many = any_nodegroup_in_hierarchy_is_cardinality_n(node.nodegroup, graph_nodes)
    expression = get_node_value_expression(node, many)
    tile_query = (
        TileModel.objects.filter(
            nodegroup_id=node.nodegroup_id,
            resourceinstance_id=OuterRef("resourceinstanceid"),
        )
        .annotate(node_value=expression)
        .values("node_value")
        .order_by("parenttile", "sortorder")
    )

    if many:
        # None is a better representation than [None] for this subquery.
        # The python representation on aliased_data will still be [].
        return NullIf(
            CardinalityNSubquery(tile_query),
            Value([None]),
        )
    else:
        return tile_query


def get_node_value_expression(node, many: bool):
    node_lookup = f"data__{node.pk}"
    factory = DataTypeFactory()
    instance = factory.get_instance(node.datatype)
    output_field = factory.get_model_field(instance)
    if node.datatype in DATATYPES_NEEDING_KEY_TEXT_TRANSFORM:
        default = KT(node_lookup)
    else:
        default = F(node_lookup)
    if node.datatype in DATATYPES_NEEDING_CAST or many:
        default = Cast(default, output_field=output_field)
    else:
        default = ExpressionWrapper(default, output_field=output_field)
    return Case(When(**{node_lookup: None}, then=Value(None)), default=default)


def get_nodegroups_here_and_below(start_nodegroup):
    accumulator = []

    def accumulate(nodegroup):
        nonlocal accumulator
        accumulator.append(nodegroup)
        # arches_version==9.0.0
        if arches_version >= (8, 0):
            children_attr = nodegroup.children
        else:
            children_attr = nodegroup.nodegroup_set
        for child_nodegroup in children_attr.prefetch_related("node_set__nodegroup"):
            accumulate(child_nodegroup)

    accumulate(start_nodegroup)
    return accumulator


def filter_nodes_by_highest_parent(nodes, aliases):
    filtered_nodes = set()
    for alias in aliases:
        for node in nodes:
            if node.alias == alias:
                break
        else:
            raise ValueError("Node alias {alias} not found in nodes.")
        nodegroups = get_nodegroups_here_and_below(node.nodegroup)
        for nodegroup in nodegroups:
            filtered_nodes |= set(nodegroup.node_set.all())

    return filtered_nodes


def any_nodegroup_in_hierarchy_is_cardinality_n(nodegroup, graph_nodes):
    # Avoid verbose prefetching by just building a lookup locally.
    parent_nodegroup_lookup = {
        node.nodegroup.parentnodegroup_id: node.nodegroup.parentnodegroup
        for node in graph_nodes
        if node.nodegroup
    }
    cardinality_n_found = False
    breaker = 0
    while not cardinality_n_found and nodegroup and breaker < 1000:
        if nodegroup.cardinality == "n":
            cardinality_n_found = True
        if nodegroup.parentnodegroup_id:
            nodegroup = parent_nodegroup_lookup[nodegroup.parentnodegroup_id]
        else:
            nodegroup = None
        breaker += 1

    return cardinality_n_found


def get_recursive_prefetches(lookup_str, *, recursive_part="children", depth):
    """
    Future: see various solutions mentioned here for avoiding
    "magic number" depth traversal (but the magic number is harmless,
    causes no additional queries beyond actual depth):
    https://forum.djangoproject.com/t/prefetching-relations-to-arbitrary-depth/39328
    """
    prefetches = []
    for i in range(1, depth + 1):
        recursive_lookup = "__".join([recursive_part] * i)
        prefetches.append(lookup_str.replace(recursive_part, recursive_lookup))
    return prefetches


def append_tiles_recursively(resource_or_tile):
    from arches_querysets.models import TileTree

    if not vars(resource_or_tile.aliased_data):
        raise RuntimeError("aliased_data is empty")

    for alias, maybe_tiles in vars(resource_or_tile.aliased_data).items():
        if maybe_tiles in (None, []):
            try:
                resource_or_tile.append_tile(alias)
            except (ValueError, RuntimeError):  # not a nodegroup or alias not found
                continue

            maybe_tiles = getattr(resource_or_tile.aliased_data, alias)
        if not isinstance(maybe_tiles, list):
            maybe_tiles = [maybe_tiles]
        for tile in maybe_tiles:
            if isinstance(tile, TileTree):
                tile.fill_blanks()


def ensure_request(request, force_admin=False):
    """Allow server-side usage when not going through middleware."""
    if not request:
        request = HttpRequest()
        username = "admin" if force_admin else "anonymous"
        request.user = (
            User.objects.filter(username=username).select_related("userprofile").get()
        )

    # arches_version==9.0.0: remove & fix call sites to check .viewable_nodegroups
    try:
        request.user.userprofile.cached_viewable_nodegroups
    except AttributeError:
        request.user.userprofile.cached_viewable_nodegroups = (
            # This property is not cached by core until Arches 8.1
            request.user.userprofile.viewable_nodegroups
        )

    return request



import uuid
from collections import defaultdict
from functools import partial

from django.conf import settings
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.db import models
from django.utils.translation import gettext as _
from rest_framework.exceptions import ValidationError
from rest_framework import renderers
from rest_framework import serializers
from rest_framework.fields import empty

from arches import VERSION as arches_version
from arches.app.models.fields.i18n import I18n_JSON, I18n_String
from arches.app.models.models import GraphModel, Node, NodeGroup
from arches.app.utils.betterJSONSerializer import JSONSerializer

from arches_querysets.datatypes.datatypes import DataTypeFactory
from arches_querysets.models import AliasedData, ResourceTileTree, TileTree
from arches_querysets.utils.models import ensure_request
from arches_querysets.rest_framework.field_mixins import NodeValueMixin
from arches_querysets.rest_framework.utils import get_nodegroup_alias_lookup


def _make_tile_serializer(
    *,
    nodegroup_alias,
    cardinality,
    sortorder,
    slug,
    graph_nodes,
    request,
    nodegroup_alias_lookup=None,
    nodes="__all__",
    exclude_children=False,
) -> type["ArchesTileSerializer"]:
    """
    DRF encourages a declarative programming style with classes. You, as
    the project implementer, can follow that style if you wish, but we've
    put some effort toward hiding this complexity from you by generating
    classes on the fly by default.
    """
    init_kwargs = {
        "required": False,
        "allow_null": False,
        "graph_nodes": graph_nodes,
        "graph_slug": slug,
        "root_node": nodegroup_alias,
        "nodegroup_alias_lookup": nodegroup_alias_lookup,
    }

    single_serializer = SingleNodegroupAliasedDataSerializer(**init_kwargs)
    multi_serializer = TileAliasedDataSerializer(**init_kwargs)

    class DynamicTileSerializer(ArchesTileSerializer):
        aliased_data = single_serializer if exclude_children else multi_serializer

        class Meta:
            model = TileTree
            graph_slug = slug
            root_node = nodegroup_alias
            fields = nodes

    name = "_".join((slug.title(), nodegroup_alias.title(), "TileSerializer"))
    serializer_class = type(name, (DynamicTileSerializer,), {})
    return serializer_class(
        many=cardinality == "n",
        required=False,
        allow_null=True,
        graph_slug=slug,
        graph_nodes=graph_nodes,
        nodegroup_alias_lookup=nodegroup_alias_lookup,
        style={"alias": nodegroup_alias, "sortorder": sortorder},
        request=request,
    )


def _wrap_serializer_field(serializer_field_class) -> type:
    """
    Ensure every serializer field in DRF's serializer_field_mapping
    resolves to one of *our* serializer fields with overrides handling
    node_value wrapping and unwrapping.
    """
    return type(
        serializer_field_class.__name__,
        (NodeValueMixin, serializer_field_class),
        {},
    )


def _fill_blank_cardinality_n_lists(serializer, data):
    """
    When fill_blanks is True, ensure that any ListSerializer field that
    is currently [] or None gets a single blank element.

    This is meant to be called from get_initial() (for /blank routes),
    but it also works for a representation dict.
    """
    if not serializer.context.get("fill_blanks") or not isinstance(data, dict):
        return data

    for field_name, field in serializer.fields.items():
        if isinstance(field, serializers.ListSerializer) and data.get(field_name) in (
            None,
            [],
        ):
            data[field_name] = [field.child.get_initial()]

    return data


def _handle_nested_aliased_data(data, *, fields_map) -> AliasedData:
    all_data = AliasedData(**data)
    for field_name, serializer in fields_map.items():
        field_data = getattr(all_data, field_name, None)
        if isinstance(serializer, ArchesTileSerializer) or (
            isinstance(serializer, serializers.ListSerializer)
            and isinstance(serializer.child, ArchesTileSerializer)
        ):
            validated_data = serializer.run_validation(field_data)

            if validated_data:
                if getattr(serializer, "many", False):
                    tile_or_tiles = [
                        TileTree(**validated_tile_data)
                        for validated_tile_data in validated_data
                    ]
                else:
                    tile_or_tiles = TileTree(**validated_data)
                setattr(all_data, field_name, tile_or_tiles)
        else:
            setattr(all_data, field_name, serializer.to_internal_value(field_data))
    return all_data


class NodeFetcherMixin:
    @property
    def graph_slug(self):
        return (
            # 1. From __init__(), e.g. TileAliasedDataSerializer
            getattr(self, "_graph_slug", None)
            # 2. From Meta options
            or self.Meta.graph_slug
            # 3. From generic view
            or self.context.get("graph_slug")
            # 4. From settings
            or getattr(settings, "SPECTACULAR_SETTINGS", {}).get(
                "GRAPH_SLUG_FOR_GENERIC_SERIALIZER"
            )
        )

    @graph_slug.setter
    def graph_slug(self, value):
        self._graph_slug = value

    @property
    def graph_nodes(self):
        if not self._graph_nodes:
            graph_nodes = []
            if getattr(self, "parent", None) and isinstance(
                self.parent, NodeFetcherMixin
            ):
                graph_nodes = self.parent.graph_nodes
            if not graph_nodes:
                graph_nodes = self._find_graph_nodes()
            self._graph_nodes = graph_nodes
        return self._graph_nodes

    @graph_nodes.setter
    def graph_nodes(self, value):
        self._graph_nodes = value

    def _find_graph_nodes(self):
        node_filters = models.Q(graph__slug=self.graph_slug, nodegroup__isnull=False)
        children = "nodegroup_set"
        # arches_version==9.0.0
        if arches_version >= (8, 0):
            node_filters &= models.Q(source_identifier=None)
            children = "children"

        return (
            Node.objects.filter(node_filters)
            .select_related("nodegroup")
            .prefetch_related(
                "nodegroup__node_set",
                "nodegroup__cardmodel_set",
                f"nodegroup__{children}",
                "cardxnodexwidget_set",
            )
        )

    @property
    def nodegroup_alias(self):
        return self.context.get("nodegroup_alias")

    def ensure_context(
        self,
        *,
        graph_slug,
        graph_nodes,
        nodegroup_alias=None,
        nodegroup_alias_lookup=None,
        request=None,
    ):
        """The view provides a context, so this is mainly here for script usage."""
        ensured_request = ensure_request(request)
        fill_blanks = ensured_request.GET.get("fill_blanks", "").lower() == "true"

        return {
            "graph_slug": graph_slug,
            "graph_nodes": graph_nodes,
            "nodegroup_alias": nodegroup_alias,
            "nodegroup_alias_lookup": (
                nodegroup_alias_lookup or get_nodegroup_alias_lookup(graph_slug)
            ),
            "request": ensured_request,
            "fill_blanks": fill_blanks,
        }


class ResourceAliasedDataSerializer(serializers.Serializer, NodeFetcherMixin):
    class Meta:
        graph_slug = None
        nodegroups = "__all__"
        fields = "__all__"
        exclude_children = False

    def __init__(self, instance=None, data=empty, **kwargs):
        super().__init__(instance, data, **kwargs)
        self.graph_nodes = kwargs.pop("graph_nodes", [])
        self._root_node_aliases = []

    def get_value(self, dictionary):
        """Avoid the branch that treats MultiPart data input as HTML."""
        return dictionary.get(self.field_name, empty)

    def get_fields(self):
        field_map = super().get_fields()
        self._root_node_aliases = []
        options = self.__class__.Meta
        if options.nodegroups == "__all__":
            only = self.context.get("nodegroup_alias")
        else:
            only = options.nodegroups

        # Create serializers for top-level nodegroups.
        for node in self.graph_nodes:
            profile = self.context["request"].user.userprofile
            if (
                not node.nodegroup_id
                # arches_version==9.0.0 replace cached_viewable_nodegroups with viewable_nodegroups
                or str(node.nodegroup_id) not in profile.cached_viewable_nodegroups
                or node.nodegroup.parentnodegroup_id
                or not node.alias  # arches_version==9.0.0: remove `or not node.alias`
            ):
                continue
            if only and node.nodegroup.grouping_node.alias not in only:
                continue
            if node.pk == node.nodegroup.pk:
                self._root_node_aliases.append(node.alias)
                if node.alias not in field_map:
                    sortorder = 0
                    if node.nodegroup.cardmodel_set.all():
                        sortorder = node.nodegroup.cardmodel_set.all()[0].sortorder
                    # TODO: check "fields" option in Meta for node level control.
                    field_map[node.alias] = _make_tile_serializer(
                        slug=self.graph_slug,
                        nodegroup_alias=node.alias,
                        cardinality=node.nodegroup.cardinality,
                        graph_nodes=self.graph_nodes,
                        nodegroup_alias_lookup=self.context["nodegroup_alias_lookup"],
                        sortorder=sortorder,
                        exclude_children=options.exclude_children,
                        request=self.context["request"],
                    )

        return field_map

    def get_default_field_names(self, declared_fields, model_info):
        field_names = super().get_default_field_names(declared_fields, model_info)
        if self.Meta.fields != "__all__":
            raise NotImplementedError  # TODO...
        if self.Meta.nodegroups == "__all__":
            field_names.extend(self._root_node_aliases)
        else:
            field_names.extend(self.Meta.nodegroups)
        return field_names

    def get_initial(self):
        """
        Used for /blank routes: build initial data and, if fill_blanks is True,
        ensure cardinality-n lists get one blank element.
        """
        initial = super().get_initial()
        return _fill_blank_cardinality_n_lists(self, initial)

    def to_internal_value(self, data):
        """Make nested aliased data writable."""
        self.initial_data = data
        return _handle_nested_aliased_data(data, fields_map=self.fields)

    def validate(self, attrs):
        aliased_data = attrs  # thanks to to_internal_value().
        if unknown_keys := set(vars(aliased_data)) - set(self.fields):
            raise ValidationError({unknown_keys.pop(): "Unexpected field"})
        return attrs


class ResourceTopNodegroupsAliasedDataSerializer(ResourceAliasedDataSerializer):
    class Meta:
        graph_slug = None
        nodegroups = "__all__"
        fields = "__all__"
        exclude_children = True


class TileAliasedDataSerializer(serializers.ModelSerializer, NodeFetcherMixin):
    serializer_field_mapping = {
        model_field: _wrap_serializer_field(serializer_field)
        for model_field, serializer_field in serializers.ModelSerializer.serializer_field_mapping.items()
    }

    class Meta:
        model = TileTree
        graph_slug = None
        # If None, supply by a route providing a <slug:nodegroup_alias> component
        root_node = None
        fields = "__all__"
        exclude_children = False

    def __init__(self, instance=None, data=empty, **kwargs):
        self.graph_slug = kwargs.pop("graph_slug", None)
        self.graph_nodes = kwargs.pop("graph_nodes", [])
        self._nodegroup_alias_lookup = kwargs.pop("nodegroup_alias_lookup", {})
        self._root_node = kwargs.pop("root_node", None)
        super().__init__(instance, data, **kwargs)
        self._child_nodegroup_aliases = []
        if not self.url_field_name:
            # Avoid cryptic errors in case a node alias collides with
            # DRF's default URL field name ("url")
            self.url_field_name = "__nonexistent__"

    def __deepcopy__(self, memo):
        ret = super().__deepcopy__(memo)
        ret._graph_slug = self._graph_slug
        ret._graph_nodes = self._graph_nodes
        return ret

    @classmethod
    def register_custom_datatype_field(cls, model_field, serializer_field):
        cls.serializer_field_mapping[model_field] = _wrap_serializer_field(
            serializer_field
        )

    def get_value(self, dictionary):
        """Avoid the branch that treats MultiPart data input as HTML."""
        return dictionary.get(self.field_name, empty)

    def get_fields(self):
        nodegroup_alias = (
            # 1. From __init__()
            getattr(self, "_root_node", None)
            # 2. From Meta options
            or self.Meta.root_node
            # 3. From generic view
            or self.context.get("nodegroup_alias")
        )
        nodes_by_node_aliases = {node.alias: node for node in self.graph_nodes}
        try:
            self._root_node = nodes_by_node_aliases.get(nodegroup_alias)
        except KeyError:
            raise RuntimeError("missing root node")
        profile = self.context["request"].user.userprofile
        # arches_version==9.0.0: cached_viewable_nodegroups -> viewable_nodegroups
        if str(self._root_node.nodegroup_id) not in profile.cached_viewable_nodegroups:
            raise PermissionError

        field_map = super().get_fields()
        self.finalize_initial_values(field_map)

        # __all__ includes children as well.
        if self.Meta.fields == "__all__" and not self.Meta.exclude_children:
            child_query = (
                self._root_node.nodegroup.children
                # arches_version==9.0.0
                if arches_version >= (8, 0)
                else self._root_node.nodegroup.nodegroup_set
            )
            flat_node_lookup = {node.pk: node for node in self.context["graph_nodes"]}
            for child_nodegroup in child_query.all():
                # arches_version==9.0.0: cached_viewable_nodegroups -> viewable_nodegroups
                if str(child_nodegroup.pk) not in profile.cached_viewable_nodegroups:
                    continue
                child_nodegroup_alias = self.context["nodegroup_alias_lookup"][
                    child_nodegroup.pk
                ]
                self._child_nodegroup_aliases.append(child_nodegroup_alias)
                child_nodegroup = flat_node_lookup[child_nodegroup.pk].nodegroup

                if (
                    child_nodegroup_alias in nodes_by_node_aliases
                    and child_nodegroup not in field_map
                ):
                    sortorder = 0
                    if child_nodegroup.cardmodel_set.all():
                        sortorder = child_nodegroup.cardmodel_set.all()[0].sortorder
                    field_map[child_nodegroup_alias] = _make_tile_serializer(
                        nodegroup_alias=child_nodegroup_alias,
                        nodegroup_alias_lookup=self.context["nodegroup_alias_lookup"],
                        cardinality=child_nodegroup.cardinality,
                        slug=self.graph_slug,
                        graph_nodes=self.graph_nodes,
                        sortorder=sortorder,
                        request=self.context["request"],
                    )

        return field_map

    def get_default_field_names(self, declared_fields, model_info):
        field_names = []
        if self.Meta.fields == "__all__":
            for sibling_node in self._root_node.nodegroup.node_set.all():
                if sibling_node.datatype != "semantic":
                    field_names.append(sibling_node.alias)

        field_names.extend(self._child_nodegroup_aliases)
        return field_names

    def get_initial(self):
        """
        Used for /blank routes: build initial data and, if fill_blanks is True,
        ensure cardinality-n lists get one blank element.
        """
        initial = super().get_initial()
        return _fill_blank_cardinality_n_lists(self, initial)

    def build_unknown_field(self, field_name, model_class):
        for node in self.graph_nodes:
            if node.alias == field_name:
                break
        else:
            raise Node.DoesNotExist(
                f"Node with alias {field_name} not found in graph {self.graph_slug}"
            )

        # arches_version==9.0.0: cached_viewable_nodegroups -> viewable_nodegroups
        if (
            node.datatype == "semantic"
            and node.nodegroup.grouping_node == node
            and str(node.nodegroup_id)
            in self.context["request"].user.userprofile.cached_viewable_nodegroups
        ):
            sortorder = 0
            if node.nodegroup.cardmodel_set.all():
                sortorder = node.nodegroup.cardmodel_set.all()[0].sortorder
            model_field = _make_tile_serializer(
                slug=self.graph_slug,
                nodegroup_alias=node.alias,
                cardinality=node.nodegroup.cardinality,
                graph_nodes=self.graph_nodes,
                sortorder=sortorder,
                request=self.context["request"],
            )
        else:
            dt_instance = DataTypeFactory().get_instance(node.datatype)
            model_field = DataTypeFactory.get_model_field(dt_instance)
        model_field.model = model_class
        model_field.blank = not node.isrequired
        try:
            cross = node.cardxnodexwidget_set.all()[0]
            label = cross.label
            visible = cross.visible
            config = cross.config
            sortorder = cross.sortorder or 0
        except (IndexError, ObjectDoesNotExist, MultipleObjectsReturned):
            label = I18n_String()
            visible = True
            config = I18n_JSON()
            sortorder = 0

        ret = self.build_standard_field(field_name, model_field)
        ret[1]["required"] = node.isrequired
        ret[1]["help_text"] = config.serialize().get("placeholder")
        ret[1]["label"] = label.serialize()
        ret[1]["style"] = {
            "alias": node.alias,
            "visible": visible,
            "widget_config": config,
            "datatype": node.datatype,
            "sortorder": sortorder,
        }
        # Default value finalized (in bulk) via finalize_initial_values().
        ret[1]["initial"] = TileTree.get_default_value(node)

        return ret

    def finalize_initial_values(self, field_map):
        """Get display values for initial values in bulk if possible."""
        nodes_by_alias = {node.alias: node for node in self.graph_nodes}

        values_by_datatype = defaultdict(list)
        for field_name, field in field_map.items():
            node = nodes_by_alias[field_name]
            values_by_datatype[node.datatype].append(field.initial)

        datatype_contexts = {}
        # Get datatype context querysets per serializer (globally would be better.)
        # Also note this is copied in TileTreeQuerySet._set_aliased_data()
        for datatype, values in values_by_datatype.items():
            datatype_instance = DataTypeFactory().get_instance(datatype)
            bulk_values = datatype_instance.get_display_value_context_in_bulk(values)
            datatype_instance.set_display_value_context_in_bulk(bulk_values)
            datatype_contexts[datatype] = bulk_values

        for field_name, field in field_map.items():
            node = nodes_by_alias[field_name]
            default_val = field.initial
            # It's a little roundabout to instantiate a tile like this, but the underlying
            # methods expect tiles in case there are provisional edits there.
            dummy_tile = TileTree(data={str(node.pk): default_val})
            pair = dummy_tile.get_value_with_context(
                node, node_value=default_val, datatype_contexts=datatype_contexts
            )
            field.initial = pair

    def to_internal_value(self, data):
        """Make nested aliased data writable."""
        self.initial_data = data
        return _handle_nested_aliased_data(data, fields_map=self.fields)

    def validate(self, attrs):
        aliased_data = attrs  # thanks to to_internal_value().
        if unknown_keys := set(vars(aliased_data)) - set(self.fields):
            raise ValidationError({unknown_keys.pop(): "Unexpected field"})

        if validate_method := getattr(
            self.root, f"validate_{self._root_node.alias}", None
        ):
            attrs = validate_method(attrs, initial_tile_data=self.parent.initial_data)

        return attrs


class SingleNodegroupAliasedDataSerializer(TileAliasedDataSerializer):
    class Meta:
        model = TileTree
        graph_slug = None
        # If None, supply by a route providing a <slug:nodegroup_alias> component
        root_node = None
        fields = "__all__"
        exclude_children = True


class ArchesTileSerializer(serializers.ModelSerializer, NodeFetcherMixin):
    # These fields are declared here in full instead of massaged via
    # "extra_kwargs" in class Meta to support subclassing by TileAliasedDataSerializer.
    tileid = serializers.UUIDField(validators=[], required=False, allow_null=True)
    resourceinstance = serializers.PrimaryKeyRelatedField(
        queryset=ResourceTileTree.objects.all(),
        required=False,
        allow_null=True,
        html_cutoff=0,
    )
    nodegroup = serializers.PrimaryKeyRelatedField(
        queryset=NodeGroup.objects.all(), required=False, allow_null=True, html_cutoff=0
    )
    parenttile = serializers.PrimaryKeyRelatedField(
        queryset=TileTree.objects.prefetch_related(None),
        required=False,
        allow_null=True,
        html_cutoff=0,
    )
    aliased_data = TileAliasedDataSerializer(required=False, allow_null=False)

    class Meta:
        model = TileTree
        # If None, supply by a route providing a <slug:graph> component
        graph_slug = None
        # If None, supply by a route providing a <slug:nodegroup_alias> component
        root_node = None
        fields = "__all__"

    def __init__(self, instance=None, data=empty, *, context=None, **kwargs):
        # TODO(next): reduce number of paths through this code
        self._graph_slug = kwargs.pop(
            "graph_slug", context.get("graph_slug", None) if context else None
        )
        self._graph_nodes = kwargs.pop(
            "graph_nodes", context.get("graph_nodes", None) if context else None
        )
        if not context:
            context = self.ensure_context(
                graph_slug=self.graph_slug,
                graph_nodes=self.graph_nodes,
                nodegroup_alias=kwargs.pop("nodegroup_alias", None),
                nodegroup_alias_lookup=kwargs.pop("nodegroup_alias_lookup", {}),
                request=kwargs.pop("request", None),
            )
        kwargs["context"] = context
        super().__init__(instance, data, **kwargs)
        self._child_nodegroup_aliases = []

    def to_representation(self, data):
        """Prevent newly minted blank tiles from serializing with pk's if the
        fill_blanks option is in force."""
        ret = super().to_representation(data)
        if self.context["fill_blanks"] and isinstance(data, TileTree):
            if data._state.adding:
                ret["tileid"] = None
                ret["parenttile"] = None
        return ret

    def to_internal_value(self, data):
        ret = super().to_internal_value(data)
        # Provide a tileid here instead of via default=uuid.uuid4 on the field definition
        # because we also have /blank routes that use this serializer.
        if ret.get("tileid", object()) is None:
            ret["tileid"] = uuid.uuid4()
        # arches_version==9.0.0
        if arches_version < (8, 0):
            # Simulate field default provided by Arches 8+.
            ret["data"] = {}
        return ret

    def get_default_field_names(self, declared_fields, model_info):
        field_names = super().get_default_field_names(declared_fields, model_info)
        try:
            field_names.remove("data")
        except ValueError:
            pass
        return field_names

    def create(self, validated_data):
        # Provide some additional context to TileTree.__init__()
        validated_data["__request"] = self.context["request"]
        validated_data["__as_representation"] = True
        graph_filters = models.Q(slug=self.graph_slug)
        # arches_version==9.0.0
        if arches_version >= (8, 0):
            graph_filters &= models.Q(source_identifier=None)
        graph = (
            GraphModel.objects.filter(graph_filters)
            .annotate(
                entry_nodegroup_id=NodeGroup.objects.filter(
                    node__graph=models.OuterRef("pk"),
                    node__alias=self.nodegroup_alias,
                ).values("pk")[:1]
            )
            .values("entry_nodegroup_id")
            .get()
        )
        validated_data["nodegroup_id"] = graph["entry_nodegroup_id"]
        return super().create(validated_data)


class ArchesSingleNodegroupSerializer(ArchesTileSerializer):
    aliased_data = SingleNodegroupAliasedDataSerializer(
        required=False, allow_null=False
    )


class ArchesResourceSerializer(serializers.ModelSerializer, NodeFetcherMixin):
    # aliased_data is a virtual field not inferred by serializers.ModelSerializer.
    aliased_data = ResourceAliasedDataSerializer(required=False, allow_null=False)

    # Custom read-only fields.
    graph_has_different_publication = serializers.SerializerMethodField()

    class Meta:
        model = ResourceTileTree
        # If None, supply by a route providing a <slug:graph> component
        graph_slug = None
        nodegroups = "__all__"
        fields = "__all__"
        # TODO (arches_version): when Arches 8.1 is the lowest supported version,
        # read_only_fields can be removed, since at that point we can depend on
        # the equivalent editable=False on the model fields (done in 8.1).
        read_only_fields = (
            "principaluser",
            "name",
            "descriptors",
            "legacyid",
            "graph_publication",
            "resource_instance_lifecycle_state",
        )
        extra_kwargs = {
            # Cancel the underlying default so it can be managed by
            # TileTreeOperation.for_new_resource
            "resourceinstanceid": {
                "initial": None,
                "default": None,
                "allow_null": True,
            },
            # TODO (arches_version): when Arches 8.0 is the lowest supported
            # version, "required": False can be removed, by relying on the
            # underlying blank=True on model.
            "graph": {"allow_null": True, "required": False},
        }

    def __init__(
        self,
        *args,
        graph_slug=None,
        graph_nodes=None,
        nodegroup_alias_lookup=None,
        context=None,
        **kwargs,
    ):
        self._graph_slug = graph_slug
        # TODO(next): simplify supplying context
        self._graph_nodes = graph_nodes or (context or {}).get("graph_nodes") or []
        if not context:
            context = self.ensure_context(
                graph_slug=self.graph_slug,
                graph_nodes=self.graph_nodes,
                nodegroup_alias_lookup=nodegroup_alias_lookup or {},
                request=kwargs.pop("request", None),
            )
        kwargs["context"] = context
        super().__init__(*args, **kwargs)

    def build_relational_field(self, field_name, relation_info):
        ret = super().build_relational_field(field_name, relation_info)
        # arches_version==9.0.0
        if field_name == "graph":
            if arches_version >= (8, 0):
                ret[1]["queryset"] = ret[1]["queryset"].filter(
                    slug=self.graph_slug, source_identifier=None
                )
            else:
                ret[1]["queryset"] = ret[1]["queryset"].filter(slug=self.graph_slug)
        return ret

    def validate(self, attrs):
        """Infer the graph if missing from the request."""
        if (
            "graph" in self.fields
            and not attrs.get("graph_id")
            and not attrs.get("graph")
        ):
            attrs["graph_id"] = self.fields["graph"].queryset.first().pk
        return attrs

    def create(self, validated_data):
        # Provide some additional context to ResourceTileTree.__init__()
        validated_data["__request"] = self.context["request"]
        validated_data["__as_representation"] = True
        return super().create(validated_data)

    def update(self, instance, validated_data):
        request = self.context.get("request")

        aliased_data = validated_data.pop("aliased_data", None)

        for attribute_name, attribute_value in validated_data.items():
            setattr(instance, attribute_name, attribute_value)

        if aliased_data is not None:
            instance.aliased_data = aliased_data

        instance.save(request=request)

        return instance

    def get_graph_has_different_publication(self, obj):
        # arches_version==9.0.0
        if arches_version < (8, 0):
            return False
        return obj.graph_publication_id and (
            obj.graph_publication_id != obj.graph.publication_id
        )


class ArchesResourceTopNodegroupsSerializer(ArchesResourceSerializer):
    aliased_data = ResourceTopNodegroupsAliasedDataSerializer(
        required=False, allow_null=False
    )


# Workaround for I18n_string fields
renderers.JSONRenderer.encoder_class = JSONSerializer
renderers.JSONOpenAPIRenderer.encoder_class = JSONSerializer



from django.urls import path

from arches_querysets.apps import ArchesQuerySetsConfig
from arches_querysets.rest_framework.generic_views import (
    ArchesResourceBlankView,
    ArchesResourceDetailView,
    ArchesResourceListCreateView,
    ArchesTileBlankView,
    ArchesTileDetailView,
    ArchesTileListCreateView,
)

app_name = ArchesQuerySetsConfig.name
urlpatterns = [
    path(
        "api/resource/<slug:graph>",
        ArchesResourceListCreateView.as_view(),
        name="api-resources",
    ),
    path(
        "api/resource/<slug:graph>/<uuid:pk>",
        ArchesResourceDetailView.as_view(),
        name="api-resource",
    ),
    path(
        "api/resource/<slug:graph>/blank",
        ArchesResourceBlankView.as_view(),
        name="api-resource-blank",
    ),
    path(
        "api/tile/<slug:graph>/<slug:nodegroup_alias>",
        ArchesTileListCreateView.as_view(),
        name="api-tiles",
    ),
    path(
        "api/tile/<slug:graph>/<slug:nodegroup_alias>/<uuid:pk>",
        ArchesTileDetailView.as_view(),
        name="api-tile",
    ),
    path(
        "api/tile/<slug:graph>/<slug:nodegroup_alias>/blank",
        ArchesTileBlankView.as_view(),
        name="api-tile-blank",
    ),
]
